# Copyright (C) 2011 by Canonical Ltd
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

import os

from breezy import merge, osutils, tests
from breezy.plugins import po_merge
from breezy.tests import features, script


class BlackboxTestPoMerger(script.TestCaseWithTransportAndScript):

    _test_needs_features = [features.msgmerge_feature]

    def setUp(self):
        super().setUp()
        self.builder = make_adduser_branch(self, 'adduser')
        # We need to install our hook as the test framework cleared it as part
        # of the initialization
        merge.Merger.hooks.install_named_hook(
            "merge_file_content", po_merge.po_merge_hook, ".po file merge")

    def test_merge_with_hook_gives_unexpected_results(self):
        # Since the conflicts in .pot are not seen *during* the merge, the .po
        # merge triggers the hook and creates no conflicts for fr.po. But the
        # .pot used is the one present in the tree *before* the merge.
        self.run_script("""\
$ brz branch adduser -rrevid:this work
2>Branched 2 revisions.
$ cd work
$ brz merge ../adduser -rrevid:other
2> M  po/adduser.pot
2> M  po/fr.po
2>Text conflict in po/adduser.pot
2>1 conflicts encountered.
""")

    def test_called_on_remerge(self):
        # Merge with no config for the hook to create the conflicts
        self.run_script("""\
$ brz branch adduser -rrevid:this work
2>Branched 2 revisions.
$ cd work
# set po_dirs to an empty list
$ brz merge ../adduser -rrevid:other -Opo_merge.po_dirs=
2> M  po/adduser.pot
2> M  po/fr.po
2>Text conflict in po/adduser.pot
2>Text conflict in po/fr.po
2>2 conflicts encountered.
""")
        # Fix the conflicts in the .pot file
        with open('po/adduser.pot', 'wb') as f:
            f.write(_Adduser['resolved_pot'])
        # Tell brz the conflict is resolved
        self.run_script("""\
$ brz resolve po/adduser.pot
2>1 conflict resolved, 1 remaining
# Use remerge to trigger the hook, we use the default config options here
$ brz remerge po/*.po
2>All changes applied successfully.
# There should be no conflicts anymore
$ brz conflicts
""")


def make_adduser_branch(test, relpath):
    """Helper for po_merge blackbox tests.

    This creates a branch containing the needed base revisions so tests can
    attempt merges and conflict resolutions.
    """
    builder = test.make_branch_builder(relpath)
    builder.start_series()
    builder.build_snapshot(None,
                           [('add', ('', b'root-id', 'directory', '')),
                            # Create empty files
                            ('add', ('po', b'dir-id', 'directory', None),),
                            ('add', ('po/adduser.pot', b'pot-id', 'file',
                                     _Adduser['base_pot'])),
                            ('add', ('po/fr.po', b'po-id', 'file',
                                     _Adduser['base_po'])),
                            ], revision_id=b'base')
    # The 'other' branch
    builder.build_snapshot([b'base'],
                           [('modify', ('po/adduser.pot',
                                        _Adduser['other_pot'])),
                            ('modify', ('po/fr.po',
                                        _Adduser['other_po'])),
                            ], revision_id=b'other')
    # The 'this' branch
    builder.build_snapshot([b'base'],
                           [('modify', ('po/adduser.pot', _Adduser['this_pot'])),
                            ('modify', ('po/fr.po', _Adduser['this_po'])),
                            ], revision_id=b'this')
    # builder.get_branch() tip is now 'this'
    builder.finish_series()
    return builder


class TestAdduserBranch(script.TestCaseWithTransportAndScript):
    """Sanity checks on the adduser branch content."""

    def setUp(self):
        super().setUp()
        self.builder = make_adduser_branch(self, 'adduser')

    def assertAdduserBranchContent(self, revid):
        env = dict(revid=revid, branch_name=revid)
        self.run_script("""\
$ brz branch adduser -rrevid:%(revid)s %(branch_name)s
""" % env, null_output_matches_anything=True)
        self.assertFileEqual(_Adduser['%(revid)s_pot' % env],
                             '%(branch_name)s/po/adduser.pot' % env)
        self.assertFileEqual(_Adduser['%(revid)s_po' % env],
                             '%(branch_name)s/po/fr.po' % env)

    def test_base(self):
        self.assertAdduserBranchContent('base')

    def test_this(self):
        self.assertAdduserBranchContent('this')

    def test_other(self):
        self.assertAdduserBranchContent('other')


# Real content from the adduser package so we don't have to guess about format
# details. This is declared at the end of the file to avoid cluttering the
# beginning of the file.

_Adduser = dict(
    base_pot=osutils.safe_utf8(r"""# SOME DESCRIPTIVE TITLE.
# Copyright (C) YEAR THE PACKAGE'S COPYRIGHT HOLDER
# This file is distributed under the same license as the PACKAGE package.
# FIRST AUTHOR <EMAIL@ADDRESS>, YEAR.
#
#, fuzzy
msgid ""
msgstr ""
"Project-Id-Version: PACKAGE VERSION\n"
"Report-Msgid-Bugs-To: adduser-devel@example.com\n"
"POT-Creation-Date: 2007-01-17 21:50+0100\n"
"PO-Revision-Date: YEAR-MO-DA HO:MI+ZONE\n"
"Last-Translator: FULL NAME <EMAIL@ADDRESS>\n"
"Language-Team: LANGUAGE <LL@example.com>\n"
"MIME-Version: 1.0\n"
"Content-Type: text/plain; charset=CHARSET\n"
"Content-Transfer-Encoding: 8bit\n"

#. everyone can issue "--help" and "--version", but only root can go on
#: ../adduser:135
msgid "Only root may add a user or group to the system.\n"
msgstr ""

#: ../adduser:188
msgid "Warning: The home dir you specified already exists.\n"
msgstr ""

"""),
    this_pot=osutils.safe_utf8(r"""# SOME DESCRIPTIVE TITLE.
# Copyright (C) YEAR THE PACKAGE'S COPYRIGHT HOLDER
# This file is distributed under the same license as the PACKAGE package.
# FIRST AUTHOR <EMAIL@ADDRESS>, YEAR.
#
#, fuzzy
msgid ""
msgstr ""
"Project-Id-Version: PACKAGE VERSION\n"
"Report-Msgid-Bugs-To: adduser-devel@example.com\n"
"POT-Creation-Date: 2011-01-06 21:06+0000\n"
"PO-Revision-Date: YEAR-MO-DA HO:MI+ZONE\n"
"Last-Translator: FULL NAME <EMAIL@ADDRESS>\n"
"Language-Team: LANGUAGE <LL@example.com>\n"
"Language: \n"
"MIME-Version: 1.0\n"
"Content-Type: text/plain; charset=CHARSET\n"
"Content-Transfer-Encoding: 8bit\n"

#. everyone can issue "--help" and "--version", but only root can go on
#: ../adduser:152
msgid "Only root may add a user or group to the system.\n"
msgstr ""

#: ../adduser:208
#, perl-format
msgid "Warning: The home dir %s you specified already exists.\n"
msgstr ""

#: ../adduser:210
#, perl-format
msgid "Warning: The home dir %s you specified can't be accessed: %s\n"
msgstr ""

"""),
    other_pot=osutils.safe_utf8(r"""# SOME DESCRIPTIVE TITLE.
# Copyright (C) YEAR THE PACKAGE'S COPYRIGHT HOLDER
# This file is distributed under the same license as the PACKAGE package.
# FIRST AUTHOR <EMAIL@ADDRESS>, YEAR.
#
#, fuzzy
msgid ""
msgstr ""
"Project-Id-Version: PACKAGE VERSION\n"
"Report-Msgid-Bugs-To: adduser-devel@example.com\n"
"POT-Creation-Date: 2010-11-21 17:13-0400\n"
"PO-Revision-Date: YEAR-MO-DA HO:MI+ZONE\n"
"Last-Translator: FULL NAME <EMAIL@ADDRESS>\n"
"Language-Team: LANGUAGE <LL@example.com>\n"
"Language: \n"
"MIME-Version: 1.0\n"
"Content-Type: text/plain; charset=CHARSET\n"
"Content-Transfer-Encoding: 8bit\n"

#. everyone can issue "--help" and "--version", but only root can go on
#: ../adduser:150
msgid "Only root may add a user or group to the system.\n"
msgstr ""

#: ../adduser:206
#, perl-format
msgid "Warning: The home dir %s you specified already exists.\n"
msgstr ""

#: ../adduser:208
#, perl-format
msgid "Warning: The home dir %s you specified can't be accessed: %s\n"
msgstr ""

"""),
    resolved_pot=osutils.safe_utf8(r"""# SOME DESCRIPTIVE TITLE.
# Copyright (C) YEAR THE PACKAGE'S COPYRIGHT HOLDER
# This file is distributed under the same license as the PACKAGE package.
# FIRST AUTHOR <EMAIL@ADDRESS>, YEAR.
#
#, fuzzy
msgid ""
msgstr ""
"Project-Id-Version: PACKAGE VERSION\n"
"Report-Msgid-Bugs-To: adduser-devel@example.com\n"
"POT-Creation-Date: 2011-10-19 12:50-0700\n"
"PO-Revision-Date: YEAR-MO-DA HO:MI+ZONE\n"
"Last-Translator: FULL NAME <EMAIL@ADDRESS>\n"
"Language-Team: LANGUAGE <LL@example.com>\n"
"Language: \n"
"MIME-Version: 1.0\n"
"Content-Type: text/plain; charset=CHARSET\n"
"Content-Transfer-Encoding: 8bit\n"

#. everyone can issue "--help" and "--version", but only root can go on
#: ../adduser:152
msgid "Only root may add a user or group to the system.\n"
msgstr ""

#: ../adduser:208
#, perl-format
msgid "Warning: The home dir %s you specified already exists.\n"
msgstr ""

#: ../adduser:210
#, perl-format
msgid "Warning: The home dir %s you specified can't be accessed: %s\n"
msgstr ""

"""),
    base_po=osutils.safe_utf8(r"""# adduser's manpages translation to French
# Copyright (C) 2004 Software in the Public Interest
# This file is distributed under the same license as the adduser package
#
# Translators:
# Jean-Baka Domelevo Entfellner <domelevo@example.com>, 2009.
#
msgid ""
msgstr ""
"Project-Id-Version: adduser 3.111\n"
"Report-Msgid-Bugs-To: adduser-devel@example.com\n"
"POT-Creation-Date: 2007-01-17 21:50+0100\n"
"PO-Revision-Date: 2010-01-21 10:36+0100\n"
"Last-Translator: Jean-Baka Domelevo Entfellner <domelevo@example.com>\n"
"Language-Team: Debian French Team <debian-l10n-french@example.com>\n"
"Language: \n"
"MIME-Version: 1.0\n"
"Content-Type: text/plain; charset=UTF-8\n"
"Content-Transfer-Encoding: 8bit\n"
"X-Poedit-Language: French\n"
"X-Poedit-Country: FRANCE\n"

# type: Plain text
#. everyone can issue "--help" and "--version", but only root can go on
#: ../adduser:135
msgid "Only root may add a user or group to the system.\n"
msgstr ""
"Seul le superutilisateur est autorisé à ajouter un utilisateur ou un groupe "
"au système.\n"

#: ../adduser:188
msgid "Warning: The home dir you specified already exists.\n"
msgstr ""
"Attention ! Le répertoire personnel que vous avez indiqué existe déjà.\n"

"""),
    this_po=osutils.safe_utf8(r"""# adduser's manpages translation to French
# Copyright (C) 2004 Software in the Public Interest
# This file is distributed under the same license as the adduser package
#
# Translators:
# Jean-Baka Domelevo Entfellner <domelevo@example.com>, 2009.
#
msgid ""
msgstr ""
"Project-Id-Version: adduser 3.111\n"
"Report-Msgid-Bugs-To: adduser-devel@example.com\n"
"POT-Creation-Date: 2010-10-12 15:48+0200\n"
"PO-Revision-Date: 2010-01-21 10:36+0100\n"
"Last-Translator: Jean-Baka Domelevo Entfellner <domelevo@example.com>\n"
"Language-Team: Debian French Team <debian-l10n-french@example.com>\n"
"Language: \n"
"MIME-Version: 1.0\n"
"Content-Type: text/plain; charset=UTF-8\n"
"Content-Transfer-Encoding: 8bit\n"
"X-Poedit-Language: French\n"
"X-Poedit-Country: FRANCE\n"

# type: Plain text
#. everyone can issue "--help" and "--version", but only root can go on
#: ../adduser:152
msgid "Only root may add a user or group to the system.\n"
msgstr ""
"Seul le superutilisateur est autorisé à ajouter un utilisateur ou un groupe "
"au système.\n"

#: ../adduser:208
#, fuzzy, perl-format
msgid "Warning: The home dir %s you specified already exists.\n"
msgstr ""
"Attention ! Le répertoire personnel que vous avez indiqué existe déjà.\n"

#: ../adduser:210
#, fuzzy, perl-format
msgid "Warning: The home dir %s you specified can't be accessed: %s\n"
msgstr ""
"Attention ! Le répertoire personnel que vous avez indiqué existe déjà.\n"

"""),
    other_po=osutils.safe_utf8(r"""# adduser's manpages translation to French
# Copyright (C) 2004 Software in the Public Interest
# This file is distributed under the same license as the adduser package
#
# Translators:
# Jean-Baka Domelevo Entfellner <domelevo@example.com>, 2009, 2010.
#
msgid ""
msgstr ""
"Project-Id-Version: adduser 3.112+nmu2\n"
"Report-Msgid-Bugs-To: adduser-devel@example.com\n"
"POT-Creation-Date: 2010-11-21 17:13-0400\n"
"PO-Revision-Date: 2010-11-10 11:08+0100\n"
"Last-Translator: Jean-Baka Domelevo-Entfellner <domelevo@example.com>\n"
"Language-Team: Debian French Team <debian-l10n-french@example.com>\n"
"Language: \n"
"MIME-Version: 1.0\n"
"Content-Type: text/plain; charset=UTF-8\n"
"Content-Transfer-Encoding: 8bit\n"
"X-Poedit-Country: FRANCE\n"

# type: Plain text
#. everyone can issue "--help" and "--version", but only root can go on
#: ../adduser:150
msgid "Only root may add a user or group to the system.\n"
msgstr ""
"Seul le superutilisateur est autorisé à ajouter un utilisateur ou un groupe "
"au système.\n"

#: ../adduser:206
#, perl-format
msgid "Warning: The home dir %s you specified already exists.\n"
msgstr ""
"Attention ! Le répertoire personnel que vous avez indiqué (%s) existe déjà.\n"

#: ../adduser:208
#, perl-format
msgid "Warning: The home dir %s you specified can't be accessed: %s\n"
msgstr ""
"Attention ! Impossible d'accéder au répertoire personnel que vous avez "
"indiqué (%s) : %s.\n"

"""),
    resolved_po=osutils.safe_utf8(r"""# adduser's manpages translation to French
# Copyright (C) 2004 Software in the Public Interest
# This file is distributed under the same license as the adduser package
#
# Translators:
# Jean-Baka Domelevo Entfellner <domelevo@example.com>, 2009, 2010.
#
msgid ""
msgstr ""
"Project-Id-Version: adduser 3.112+nmu2\n"
"Report-Msgid-Bugs-To: adduser-devel@example.com\n"
"POT-Creation-Date: 2011-10-19 12:50-0700\n"
"PO-Revision-Date: 2010-11-10 11:08+0100\n"
"Last-Translator: Jean-Baka Domelevo-Entfellner <domelevo@example.com>\n"
"Language-Team: Debian French Team <debian-l10n-french@example.com>\n"
"Language: \n"
"MIME-Version: 1.0\n"
"Content-Type: text/plain; charset=UTF-8\n"
"Content-Transfer-Encoding: 8bit\n"
"X-Poedit-Country: FRANCE\n"

# type: Plain text
#. everyone can issue "--help" and "--version", but only root can go on
#: ../adduser:152
msgid "Only root may add a user or group to the system.\n"
msgstr ""
"Seul le superutilisateur est autorisé à ajouter un utilisateur ou un groupe "
"au système.\n"

#: ../adduser:208
#, perl-format
msgid "Warning: The home dir %s you specified already exists.\n"
msgstr ""
"Attention ! Le répertoire personnel que vous avez indiqué (%s) existe déjà.\n"

#: ../adduser:210
#, perl-format
msgid "Warning: The home dir %s you specified can't be accessed: %s\n"
msgstr ""
"Attention ! Impossible d'accéder au répertoire personnel que vous avez "
"indiqué (%s) : %s.\n"

"""),
)



# Copyright (C) 2006-2012, 2016 Canonical Ltd
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

"""Tests for the Repository facility that are not interface tests.

For interface tests see tests/per_repository/*.py.

For concrete class tests see this file, and for storage formats tests
also see this file.
"""

from stat import S_ISDIR

import breezy
from breezy import controldir, errors, osutils, repository
from breezy import revision as _mod_revision
from breezy import tests, transport, upgrade, workingtree
from breezy.bzr import (btree_index, bzrdir, groupcompress_repo, inventory,
                        knitpack_repo, knitrepo, pack_repo)
from breezy.bzr import repository as bzrrepository
from breezy.bzr import versionedfile, vf_repository, vf_search
from breezy.bzr.btree_index import BTreeBuilder, BTreeGraphIndex
from breezy.bzr.index import GraphIndex
from breezy.errors import UnknownFormatError
from breezy.repository import RepositoryFormat
from breezy.tests import TestCase, TestCaseWithTransport


class TestDefaultFormat(TestCase):

    def test_get_set_default_format(self):
        old_default = controldir.format_registry.get('default')
        old_default_help = controldir.format_registry.get_help('default')
        private_default = old_default().repository_format.__class__
        old_format = repository.format_registry.get_default()
        self.assertTrue(isinstance(old_format, private_default))

        def make_sample_bzrdir():
            my_bzrdir = bzrdir.BzrDirMetaFormat1()
            my_bzrdir.repository_format = SampleRepositoryFormat()
            return my_bzrdir
        controldir.format_registry.remove('default')
        controldir.format_registry.register('sample', make_sample_bzrdir, '')
        controldir.format_registry.set_default('sample')
        # creating a repository should now create an instrumented dir.
        try:
            # the default branch format is used by the meta dir format
            # which is not the default bzrdir format at this point
            dir = bzrdir.BzrDirMetaFormat1().initialize('memory:///')
            result = dir.create_repository()
            self.assertEqual(result, 'A bzr repository dir')
        finally:
            controldir.format_registry.remove('default')
            controldir.format_registry.remove('sample')
            controldir.format_registry.register(
                'default', old_default, old_default_help)
        self.assertIsInstance(repository.format_registry.get_default(),
                              old_format.__class__)


class SampleRepositoryFormat(bzrrepository.RepositoryFormatMetaDir):
    """A sample format

    this format is initializable, unsupported to aid in testing the
    open and open(unsupported=True) routines.
    """

    @classmethod
    def get_format_string(cls):
        """See RepositoryFormat.get_format_string()."""
        return b"Sample .bzr repository format."

    def initialize(self, a_controldir, shared=False):
        """Initialize a repository in a BzrDir"""
        t = a_controldir.get_repository_transport(self)
        t.put_bytes('format', self.get_format_string())
        return 'A bzr repository dir'

    def is_supported(self):
        return False

    def open(self, a_controldir, _found=False):
        return "opened repository."


class SampleExtraRepositoryFormat(repository.RepositoryFormat):
    """A sample format that can not be used in a metadir

    """

    def get_format_string(self):
        raise NotImplementedError


class TestRepositoryFormat(TestCaseWithTransport):
    """Tests for the Repository format detection used by the bzr meta dir facility.BzrBranchFormat facility."""

    def test_find_format(self):
        # is the right format object found for a repository?
        # create a branch with a few known format objects.
        # this is not quite the same as
        self.build_tree(["foo/", "bar/"])

        def check_format(format, url):
            dir = format._matchingcontroldir.initialize(url)
            format.initialize(dir)
            found_format = bzrrepository.RepositoryFormatMetaDir.find_format(
                dir)
            self.assertIsInstance(found_format, format.__class__)
        check_format(repository.format_registry.get_default(), "bar")

    def test_find_format_no_repository(self):
        dir = bzrdir.BzrDirMetaFormat1().initialize(self.get_url())
        self.assertRaises(errors.NoRepositoryPresent,
                          bzrrepository.RepositoryFormatMetaDir.find_format,
                          dir)

    def test_from_string(self):
        self.assertIsInstance(
            SampleRepositoryFormat.from_string(
                b"Sample .bzr repository format."),
            SampleRepositoryFormat)
        self.assertRaises(AssertionError,
                          SampleRepositoryFormat.from_string,
                          b"Different .bzr repository format.")

    def test_find_format_unknown_format(self):
        dir = bzrdir.BzrDirMetaFormat1().initialize(self.get_url())
        SampleRepositoryFormat().initialize(dir)
        self.assertRaises(UnknownFormatError,
                          bzrrepository.RepositoryFormatMetaDir.find_format,
                          dir)

    def test_find_format_with_features(self):
        tree = self.make_branch_and_tree('.', format='2a')
        tree.branch.repository.update_feature_flags({b"name": b"necessity"})
        found_format = bzrrepository.RepositoryFormatMetaDir.find_format(
            tree.controldir)
        self.assertIsInstance(
            found_format, bzrrepository.RepositoryFormatMetaDir)
        self.assertEqual(found_format.features.get(b"name"), b"necessity")
        self.assertRaises(
            bzrdir.MissingFeature, found_format.check_support_status, True)
        self.addCleanup(
            bzrrepository.RepositoryFormatMetaDir.unregister_feature, b"name")
        bzrrepository.RepositoryFormatMetaDir.register_feature(b"name")
        found_format.check_support_status(True)


class TestRepositoryFormatRegistry(TestCase):

    def setUp(self):
        super().setUp()
        self.registry = repository.RepositoryFormatRegistry()

    def test_register_unregister_format(self):
        format = SampleRepositoryFormat()
        self.registry.register(format)
        self.assertEqual(format, self.registry.get(
            b"Sample .bzr repository format."))
        self.registry.remove(format)
        self.assertRaises(KeyError, self.registry.get,
                          b"Sample .bzr repository format.")

    def test_get_all(self):
        format = SampleRepositoryFormat()
        self.assertEqual([], self.registry._get_all())
        self.registry.register(format)
        self.assertEqual([format], self.registry._get_all())

    def test_register_extra(self):
        format = SampleExtraRepositoryFormat()
        self.assertEqual([], self.registry._get_all())
        self.registry.register_extra(format)
        self.assertEqual([format], self.registry._get_all())

    def test_register_extra_lazy(self):
        self.assertEqual([], self.registry._get_all())
        self.registry.register_extra_lazy(__name__,
                                          "SampleExtraRepositoryFormat")
        formats = self.registry._get_all()
        self.assertEqual(1, len(formats))
        self.assertIsInstance(formats[0], SampleExtraRepositoryFormat)


class TestFormatKnit1(TestCaseWithTransport):

    def test_attribute__fetch_order(self):
        """Knits need topological data insertion."""
        repo = self.make_repository(
            '.', format=controldir.format_registry.get('knit')())
        self.assertEqual('topological', repo._format._fetch_order)

    def test_attribute__fetch_uses_deltas(self):
        """Knits reuse deltas."""
        repo = self.make_repository(
            '.', format=controldir.format_registry.get('knit')())
        self.assertEqual(True, repo._format._fetch_uses_deltas)

    def test_disk_layout(self):
        control = bzrdir.BzrDirMetaFormat1().initialize(self.get_url())
        repo = knitrepo.RepositoryFormatKnit1().initialize(control)
        # in case of side effects of locking.
        repo.lock_write()
        repo.unlock()
        # we want:
        # format 'Bazaar-NG Knit Repository Format 1'
        # lock: is a directory
        # inventory.weave == empty_weave
        # empty revision-store directory
        # empty weaves directory
        t = control.get_repository_transport(None)
        with t.get('format') as f:
            self.assertEqualDiff(b'Bazaar-NG Knit Repository Format 1',
                                 f.read())
        # XXX: no locks left when unlocked at the moment
        # self.assertEqualDiff('', t.get('lock').read())
        self.assertTrue(S_ISDIR(t.stat('knits').st_mode))
        self.check_knits(t)
        # Check per-file knits.
        control.create_branch()
        tree = control.create_workingtree()
        tree.add(['foo'], ['file'], [b'Nasty-IdC:'])
        tree.put_file_bytes_non_atomic('foo', b'')
        tree.commit('1st post', rev_id=b'foo')
        self.assertHasKnit(t, 'knits/e8/%254easty-%2549d%2543%253a',
                           b'\nfoo fulltext 0 81  :')

    def assertHasKnit(self, t, knit_name, extra_content=b''):
        """Assert that knit_name exists on t."""
        with t.get(knit_name + '.kndx') as f:
            self.assertEqualDiff(b'# bzr knit index 8\n' + extra_content,
                                 f.read())

    def check_knits(self, t):
        """check knit content for a repository."""
        self.assertHasKnit(t, 'inventory')
        self.assertHasKnit(t, 'revisions')
        self.assertHasKnit(t, 'signatures')

    def test_shared_disk_layout(self):
        control = bzrdir.BzrDirMetaFormat1().initialize(self.get_url())
        knitrepo.RepositoryFormatKnit1().initialize(control, shared=True)
        # we want:
        # format 'Bazaar-NG Knit Repository Format 1'
        # lock: is a directory
        # inventory.weave == empty_weave
        # empty revision-store directory
        # empty weaves directory
        # a 'shared-storage' marker file.
        t = control.get_repository_transport(None)
        with t.get('format') as f:
            self.assertEqualDiff(b'Bazaar-NG Knit Repository Format 1',
                                 f.read())
        # XXX: no locks left when unlocked at the moment
        # self.assertEqualDiff('', t.get('lock').read())
        with t.get('shared-storage') as f:
            self.assertEqualDiff(b'', f.read())
        self.assertTrue(S_ISDIR(t.stat('knits').st_mode))
        self.check_knits(t)

    def test_shared_no_tree_disk_layout(self):
        control = bzrdir.BzrDirMetaFormat1().initialize(self.get_url())
        repo = knitrepo.RepositoryFormatKnit1().initialize(
            control, shared=True)
        repo.set_make_working_trees(False)
        # we want:
        # format 'Bazaar-NG Knit Repository Format 1'
        # lock ''
        # inventory.weave == empty_weave
        # empty revision-store directory
        # empty weaves directory
        # a 'shared-storage' marker file.
        t = control.get_repository_transport(None)
        with t.get('format') as f:
            self.assertEqualDiff(b'Bazaar-NG Knit Repository Format 1',
                                 f.read())
        # XXX: no locks left when unlocked at the moment
        # self.assertEqualDiff('', t.get('lock').read())
        with t.get('shared-storage') as f:
            self.assertEqualDiff(b'', f.read())
        with t.get('no-working-trees') as f:
            self.assertEqualDiff(b'', f.read())
        repo.set_make_working_trees(True)
        self.assertFalse(t.has('no-working-trees'))
        self.assertTrue(S_ISDIR(t.stat('knits').st_mode))
        self.check_knits(t)

    def test_deserialise_sets_root_revision(self):
        """We must have a inventory.root.revision

        Old versions of the XML5 serializer did not set the revision_id for
        the whole inventory. So we grab the one from the expected text. Which
        is valid when the api is not being abused.
        """
        repo = self.make_repository(
            '.', format=controldir.format_registry.get('knit')())
        inv_xml = b'<inventory format="5">\n</inventory>\n'
        inv = repo._deserialise_inventory(b'test-rev-id', [inv_xml])
        self.assertEqual(b'test-rev-id', inv.root.revision)

    def test_deserialise_uses_global_revision_id(self):
        """If it is set, then we re-use the global revision id"""
        repo = self.make_repository(
            '.', format=controldir.format_registry.get('knit')())
        inv_xml = (b'<inventory format="5" revision_id="other-rev-id">\n'
                   b'</inventory>\n')
        # Arguably, the deserialise_inventory should detect a mismatch, and
        # raise an error, rather than silently using one revision_id over the
        # other.
        self.assertRaises(AssertionError, repo._deserialise_inventory,
                          b'test-rev-id', [inv_xml])
        inv = repo._deserialise_inventory(b'other-rev-id', [inv_xml])
        self.assertEqual(b'other-rev-id', inv.root.revision)

    def test_supports_external_lookups(self):
        repo = self.make_repository(
            '.', format=controldir.format_registry.get('knit')())
        self.assertFalse(repo._format.supports_external_lookups)


class DummyRepository:
    """A dummy repository for testing."""

    _format = None
    _serializer = None

    def supports_rich_root(self):
        if self._format is not None:
            return self._format.rich_root_data
        return False

    def get_graph(self):
        raise NotImplementedError

    def get_parent_map(self, revision_ids):
        raise NotImplementedError


class InterDummy(repository.InterRepository):
    """An inter-repository optimised code path for DummyRepository.

    This is for use during testing where we use DummyRepository as repositories
    so that none of the default regsitered inter-repository classes will
    MATCH.
    """

    @staticmethod
    def is_compatible(repo_source, repo_target):
        """InterDummy is compatible with DummyRepository."""
        return (isinstance(repo_source, DummyRepository) and
                isinstance(repo_target, DummyRepository))


class TestInterRepository(TestCaseWithTransport):

    def test_get_default_inter_repository(self):
        # test that the InterRepository.get(repo_a, repo_b) probes
        # for a inter_repo class where is_compatible(repo_a, repo_b) returns
        # true and returns a default inter_repo otherwise.
        # This also tests that the default registered optimised interrepository
        # classes do not barf inappropriately when a surprising repository type
        # is handed to them.
        dummy_a = DummyRepository()
        dummy_a._format = RepositoryFormat()
        dummy_a._format.supports_full_versioned_files = True
        dummy_a._format.rich_root_data = True
        dummy_b = DummyRepository()
        dummy_b._format = RepositoryFormat()
        dummy_b._format.supports_full_versioned_files = True
        dummy_b._format.rich_root_data = True
        self.assertGetsDefaultInterRepository(dummy_a, dummy_b)

    def assertGetsDefaultInterRepository(self, repo_a, repo_b):
        """Asserts that InterRepository.get(repo_a, repo_b) -> the default.

        The effective default is now InterSameDataRepository because there is
        no actual sane default in the presence of incompatible data models.
        """
        inter_repo = repository.InterRepository.get(repo_a, repo_b)
        self.assertEqual(vf_repository.InterSameDataRepository,
                         inter_repo.__class__)
        self.assertEqual(repo_a, inter_repo.source)
        self.assertEqual(repo_b, inter_repo.target)

    def test_register_inter_repository_class(self):
        # test that a optimised code path provider - a
        # InterRepository subclass can be registered and unregistered
        # and that it is correctly selected when given a repository
        # pair that it returns true on for the is_compatible static method
        # check
        dummy_a = DummyRepository()
        dummy_a._format = RepositoryFormat()
        dummy_b = DummyRepository()
        dummy_b._format = RepositoryFormat()
        repo = self.make_repository('.')
        # hack dummies to look like repo somewhat.
        dummy_a._serializer = repo._serializer
        dummy_a._format.supports_tree_reference = (
            repo._format.supports_tree_reference)
        dummy_a._format.rich_root_data = repo._format.rich_root_data
        dummy_a._format.supports_full_versioned_files = (
            repo._format.supports_full_versioned_files)
        dummy_b._serializer = repo._serializer
        dummy_b._format.supports_tree_reference = (
            repo._format.supports_tree_reference)
        dummy_b._format.rich_root_data = repo._format.rich_root_data
        dummy_b._format.supports_full_versioned_files = (
            repo._format.supports_full_versioned_files)
        repository.InterRepository.register_optimiser(InterDummy)
        try:
            # we should get the default for something InterDummy returns False
            # to
            self.assertFalse(InterDummy.is_compatible(dummy_a, repo))
            self.assertGetsDefaultInterRepository(dummy_a, repo)
            # and we should get an InterDummy for a pair it 'likes'
            self.assertTrue(InterDummy.is_compatible(dummy_a, dummy_b))
            inter_repo = repository.InterRepository.get(dummy_a, dummy_b)
            self.assertEqual(InterDummy, inter_repo.__class__)
            self.assertEqual(dummy_a, inter_repo.source)
            self.assertEqual(dummy_b, inter_repo.target)
        finally:
            repository.InterRepository.unregister_optimiser(InterDummy)
        # now we should get the default InterRepository object again.
        self.assertGetsDefaultInterRepository(dummy_a, dummy_b)


class TestRepositoryFormat1(knitrepo.RepositoryFormatKnit1):

    @classmethod
    def get_format_string(cls):
        return b"Test Format 1"


class TestRepositoryFormat2(knitrepo.RepositoryFormatKnit1):

    @classmethod
    def get_format_string(cls):
        return b"Test Format 2"


class TestRepositoryConverter(TestCaseWithTransport):

    def test_convert_empty(self):
        source_format = TestRepositoryFormat1()
        target_format = TestRepositoryFormat2()
        repository.format_registry.register(source_format)
        self.addCleanup(repository.format_registry.remove,
                        source_format)
        repository.format_registry.register(target_format)
        self.addCleanup(repository.format_registry.remove,
                        target_format)
        t = self.get_transport()
        t.mkdir('repository')
        repo_dir = bzrdir.BzrDirMetaFormat1().initialize('repository')
        repo = TestRepositoryFormat1().initialize(repo_dir)
        converter = repository.CopyConverter(target_format)
        with breezy.ui.ui_factory.nested_progress_bar() as pb:
            converter.convert(repo, pb)
        repo = repo_dir.open_repository()
        self.assertTrue(isinstance(target_format, repo._format.__class__))


class TestRepositoryFormatKnit3(TestCaseWithTransport):

    def test_attribute__fetch_order(self):
        """Knits need topological data insertion."""
        format = bzrdir.BzrDirMetaFormat1()
        format.repository_format = knitrepo.RepositoryFormatKnit3()
        repo = self.make_repository('.', format=format)
        self.assertEqual('topological', repo._format._fetch_order)

    def test_attribute__fetch_uses_deltas(self):
        """Knits reuse deltas."""
        format = bzrdir.BzrDirMetaFormat1()
        format.repository_format = knitrepo.RepositoryFormatKnit3()
        repo = self.make_repository('.', format=format)
        self.assertEqual(True, repo._format._fetch_uses_deltas)

    def test_convert(self):
        """Ensure the upgrade adds weaves for roots"""
        format = bzrdir.BzrDirMetaFormat1()
        format.repository_format = knitrepo.RepositoryFormatKnit1()
        tree = self.make_branch_and_tree('.', format)
        tree.commit("Dull commit", rev_id=b"dull")
        revision_tree = tree.branch.repository.revision_tree(b'dull')
        with revision_tree.lock_read():
            self.assertRaises(
                transport.NoSuchFile, revision_tree.get_file_lines, '')
        format = bzrdir.BzrDirMetaFormat1()
        format.repository_format = knitrepo.RepositoryFormatKnit3()
        upgrade.Convert('.', format)
        tree = workingtree.WorkingTree.open('.')
        revision_tree = tree.branch.repository.revision_tree(b'dull')
        with revision_tree.lock_read():
            revision_tree.get_file_lines('')
        tree.commit("Another dull commit", rev_id=b'dull2')
        revision_tree = tree.branch.repository.revision_tree(b'dull2')
        revision_tree.lock_read()
        self.addCleanup(revision_tree.unlock)
        self.assertEqual(b'dull', revision_tree.get_file_revision(''))

    def test_supports_external_lookups(self):
        format = bzrdir.BzrDirMetaFormat1()
        format.repository_format = knitrepo.RepositoryFormatKnit3()
        repo = self.make_repository('.', format=format)
        self.assertFalse(repo._format.supports_external_lookups)


class Test2a(tests.TestCaseWithMemoryTransport):

    def test_chk_bytes_uses_custom_btree_parser(self):
        mt = self.make_branch_and_memory_tree('test', format='2a')
        mt.lock_write()
        self.addCleanup(mt.unlock)
        mt.add([''], [b'root-id'])
        mt.commit('first')
        index = mt.branch.repository.chk_bytes._index._graph_index._indices[0]
        self.assertEqual(btree_index._gcchk_factory, index._leaf_factory)
        # It should also work if we re-open the repo
        repo = mt.branch.repository.controldir.open_repository()
        repo.lock_read()
        self.addCleanup(repo.unlock)
        index = repo.chk_bytes._index._graph_index._indices[0]
        self.assertEqual(btree_index._gcchk_factory, index._leaf_factory)

    def test_fetch_combines_groups(self):
        builder = self.make_branch_builder('source', format='2a')
        builder.start_series()
        builder.build_snapshot(None, [
            ('add', ('', b'root-id', 'directory', '')),
            ('add', ('file', b'file-id', 'file', b'content\n'))],
            revision_id=b'1')
        builder.build_snapshot([b'1'], [
            ('modify', ('file', b'content-2\n'))],
            revision_id=b'2')
        builder.finish_series()
        source = builder.get_branch()
        target = self.make_repository('target', format='2a')
        target.fetch(source.repository)
        target.lock_read()
        self.addCleanup(target.unlock)
        details = target.texts._index.get_build_details(
            [(b'file-id', b'1',), (b'file-id', b'2',)])
        file_1_details = details[(b'file-id', b'1')]
        file_2_details = details[(b'file-id', b'2')]
        # The index, and what to read off disk, should be the same for both
        # versions of the file.
        self.assertEqual(file_1_details[0][:3], file_2_details[0][:3])

    def test_format_pack_compresses_True(self):
        repo = self.make_repository('repo', format='2a')
        self.assertTrue(repo._format.pack_compresses)

    def test_inventories_use_chk_map_with_parent_base_dict(self):
        tree = self.make_branch_and_memory_tree('repo', format="2a")
        tree.lock_write()
        tree.add([''], ids=[b'TREE_ROOT'])
        revid = tree.commit("foo")
        tree.unlock()
        tree.lock_read()
        self.addCleanup(tree.unlock)
        inv = tree.branch.repository.get_inventory(revid)
        self.assertNotEqual(None, inv.parent_id_basename_to_file_id)
        inv.parent_id_basename_to_file_id._ensure_root()
        inv.id_to_entry._ensure_root()
        self.assertEqual(65536, inv.id_to_entry._root_node.maximum_size)
        self.assertEqual(
            65536, inv.parent_id_basename_to_file_id._root_node.maximum_size)

    def test_autopack_unchanged_chk_nodes(self):
        # at 20 unchanged commits, chk pages are packed that are split into
        # two groups such that the new pack being made doesn't have all its
        # pages in the source packs (though they are in the repository).
        # Use a memory backed repository, we don't need to hit disk for this
        tree = self.make_branch_and_memory_tree('tree', format='2a')
        tree.lock_write()
        self.addCleanup(tree.unlock)
        tree.add([''], ids=[b'TREE_ROOT'])
        for pos in range(20):
            tree.commit(str(pos))

    def test_pack_with_hint(self):
        tree = self.make_branch_and_memory_tree('tree', format='2a')
        tree.lock_write()
        self.addCleanup(tree.unlock)
        tree.add([''], ids=[b'TREE_ROOT'])
        # 1 commit to leave untouched
        tree.commit('1')
        to_keep = tree.branch.repository._pack_collection.names()
        # 2 to combine
        tree.commit('2')
        tree.commit('3')
        all = tree.branch.repository._pack_collection.names()
        combine = list(set(all) - set(to_keep))
        self.assertLength(3, all)
        self.assertLength(2, combine)
        tree.branch.repository.pack(hint=combine)
        final = tree.branch.repository._pack_collection.names()
        self.assertLength(2, final)
        self.assertFalse(combine[0] in final)
        self.assertFalse(combine[1] in final)
        self.assertSubset(to_keep, final)

    def test_stream_source_to_gc(self):
        source = self.make_repository('source', format='2a')
        target = self.make_repository('target', format='2a')
        stream = source._get_source(target._format)
        self.assertIsInstance(stream, groupcompress_repo.GroupCHKStreamSource)

    def test_stream_source_to_non_gc(self):
        source = self.make_repository('source', format='2a')
        target = self.make_repository('target', format='rich-root-pack')
        stream = source._get_source(target._format)
        # We don't want the child GroupCHKStreamSource
        self.assertIs(type(stream), vf_repository.StreamSource)

    def test_get_stream_for_missing_keys_includes_all_chk_refs(self):
        source_builder = self.make_branch_builder('source',
                                                  format='2a')
        # We have to build a fairly large tree, so that we are sure the chk
        # pages will have split into multiple pages.
        entries = [('add', ('', b'a-root-id', 'directory', None))]
        for i in 'abcdefghijklmnopqrstuvwxyz123456789':
            for j in 'abcdefghijklmnopqrstuvwxyz123456789':
                fname = i + j
                fid = fname.encode('utf-8') + b'-id'
                content = b'content for %s\n' % (fname.encode('utf-8'),)
                entries.append(('add', (fname, fid, 'file', content)))
        source_builder.start_series()
        source_builder.build_snapshot(None, entries, revision_id=b'rev-1')
        # Now change a few of them, so we get a few new pages for the second
        # revision
        source_builder.build_snapshot([b'rev-1'], [
            ('modify', ('aa', b'new content for aa-id\n')),
            ('modify', ('cc', b'new content for cc-id\n')),
            ('modify', ('zz', b'new content for zz-id\n')),
            ], revision_id=b'rev-2')
        source_builder.finish_series()
        source_branch = source_builder.get_branch()
        source_branch.lock_read()
        self.addCleanup(source_branch.unlock)
        target = self.make_repository('target', format='2a')
        source = source_branch.repository._get_source(target._format)
        self.assertIsInstance(source, groupcompress_repo.GroupCHKStreamSource)

        # On a regular pass, getting the inventories and chk pages for rev-2
        # would only get the newly created chk pages
        search = vf_search.SearchResult({b'rev-2'}, {b'rev-1'}, 1,
                                        {b'rev-2'})
        simple_chk_records = set()
        for vf_name, substream in source.get_stream(search):
            if vf_name == 'chk_bytes':
                for record in substream:
                    simple_chk_records.add(record.key)
            else:
                for _ in substream:
                    continue
        # 3 pages, the root (InternalNode), + 2 pages which actually changed
        self.assertEqual({(b'sha1:91481f539e802c76542ea5e4c83ad416bf219f73',),
                          (b'sha1:4ff91971043668583985aec83f4f0ab10a907d3f',),
                          (b'sha1:81e7324507c5ca132eedaf2d8414ee4bb2226187',),
                          (b'sha1:b101b7da280596c71a4540e9a1eeba8045985ee0',)},
                         set(simple_chk_records))
        # Now, when we do a similar call using 'get_stream_for_missing_keys'
        # we should get a much larger set of pages.
        missing = [('inventories', b'rev-2')]
        full_chk_records = set()
        for vf_name, substream in source.get_stream_for_missing_keys(missing):
            if vf_name == 'inventories':
                for record in substream:
                    self.assertEqual((b'rev-2',), record.key)
            elif vf_name == 'chk_bytes':
                for record in substream:
                    full_chk_records.add(record.key)
            else:
                self.fail('Should not be getting a stream of {}'.format(vf_name))
        # We have 257 records now. This is because we have 1 root page, and 256
        # leaf pages in a complete listing.
        self.assertEqual(257, len(full_chk_records))
        self.assertSubset(simple_chk_records, full_chk_records)

    def test_inconsistency_fatal(self):
        repo = self.make_repository('repo', format='2a')
        self.assertTrue(repo.revisions._index._inconsistency_fatal)
        self.assertFalse(repo.texts._index._inconsistency_fatal)
        self.assertFalse(repo.inventories._index._inconsistency_fatal)
        self.assertFalse(repo.signatures._index._inconsistency_fatal)
        self.assertFalse(repo.chk_bytes._index._inconsistency_fatal)


class TestKnitPackStreamSource(tests.TestCaseWithMemoryTransport):

    def test_source_to_exact_pack_092(self):
        source = self.make_repository('source', format='pack-0.92')
        target = self.make_repository('target', format='pack-0.92')
        stream_source = source._get_source(target._format)
        self.assertIsInstance(
            stream_source, knitpack_repo.KnitPackStreamSource)

    def test_source_to_exact_pack_rich_root_pack(self):
        source = self.make_repository('source', format='rich-root-pack')
        target = self.make_repository('target', format='rich-root-pack')
        stream_source = source._get_source(target._format)
        self.assertIsInstance(
            stream_source, knitpack_repo.KnitPackStreamSource)

    def test_source_to_exact_pack_19(self):
        source = self.make_repository('source', format='1.9')
        target = self.make_repository('target', format='1.9')
        stream_source = source._get_source(target._format)
        self.assertIsInstance(
            stream_source, knitpack_repo.KnitPackStreamSource)

    def test_source_to_exact_pack_19_rich_root(self):
        source = self.make_repository('source', format='1.9-rich-root')
        target = self.make_repository('target', format='1.9-rich-root')
        stream_source = source._get_source(target._format)
        self.assertIsInstance(
            stream_source, knitpack_repo.KnitPackStreamSource)

    def test_source_to_remote_exact_pack_19(self):
        trans = self.make_smart_server('target')
        trans.ensure_base()
        source = self.make_repository('source', format='1.9')
        target = self.make_repository('target', format='1.9')
        target = repository.Repository.open(trans.base)
        stream_source = source._get_source(target._format)
        self.assertIsInstance(
            stream_source, knitpack_repo.KnitPackStreamSource)

    def test_stream_source_to_non_exact(self):
        source = self.make_repository('source', format='pack-0.92')
        target = self.make_repository('target', format='1.9')
        stream = source._get_source(target._format)
        self.assertIs(type(stream), vf_repository.StreamSource)

    def test_stream_source_to_non_exact_rich_root(self):
        source = self.make_repository('source', format='1.9')
        target = self.make_repository('target', format='1.9-rich-root')
        stream = source._get_source(target._format)
        self.assertIs(type(stream), vf_repository.StreamSource)

    def test_source_to_remote_non_exact_pack_19(self):
        trans = self.make_smart_server('target')
        trans.ensure_base()
        source = self.make_repository('source', format='1.9')
        target = self.make_repository('target', format='1.6')
        target = repository.Repository.open(trans.base)
        stream_source = source._get_source(target._format)
        self.assertIs(type(stream_source), vf_repository.StreamSource)

    def test_stream_source_to_knit(self):
        source = self.make_repository('source', format='pack-0.92')
        target = self.make_repository('target', format='dirstate')
        stream = source._get_source(target._format)
        self.assertIs(type(stream), vf_repository.StreamSource)


class TestDevelopment6FindParentIdsOfRevisions(TestCaseWithTransport):
    """Tests for _find_parent_ids_of_revisions."""

    def setUp(self):
        super().setUp()
        self.builder = self.make_branch_builder('source')
        self.builder.start_series()
        self.builder.build_snapshot(
            None,
            [('add', ('', b'tree-root', 'directory', None))],
            revision_id=b'initial')
        self.repo = self.builder.get_branch().repository
        self.addCleanup(self.builder.finish_series)

    def assertParentIds(self, expected_result, rev_set):
        self.assertEqual(
            sorted(expected_result),
            sorted(self.repo._find_parent_ids_of_revisions(rev_set)))

    def test_simple(self):
        self.builder.build_snapshot(None, [], revision_id=b'revid1')
        self.builder.build_snapshot([b'revid1'], [], revision_id=b'revid2')
        rev_set = [b'revid2']
        self.assertParentIds([b'revid1'], rev_set)

    def test_not_first_parent(self):
        self.builder.build_snapshot(None, [], revision_id=b'revid1')
        self.builder.build_snapshot([b'revid1'], [], revision_id=b'revid2')
        self.builder.build_snapshot([b'revid2'], [], revision_id=b'revid3')
        rev_set = [b'revid3', b'revid2']
        self.assertParentIds([b'revid1'], rev_set)

    def test_not_null(self):
        rev_set = [b'initial']
        self.assertParentIds([], rev_set)

    def test_not_null_set(self):
        self.builder.build_snapshot(None, [], revision_id=b'revid1')
        rev_set = [_mod_revision.NULL_REVISION]
        self.assertParentIds([], rev_set)

    def test_ghost(self):
        self.builder.build_snapshot(None, [], revision_id=b'revid1')
        rev_set = [b'ghost', b'revid1']
        self.assertParentIds([b'initial'], rev_set)

    def test_ghost_parent(self):
        self.builder.build_snapshot(None, [], revision_id=b'revid1')
        self.builder.build_snapshot(
            [b'revid1', b'ghost'], [], revision_id=b'revid2')
        rev_set = [b'revid2', b'revid1']
        self.assertParentIds([b'ghost', b'initial'], rev_set)

    def test_righthand_parent(self):
        self.builder.build_snapshot(None, [], revision_id=b'revid1')
        self.builder.build_snapshot([b'revid1'], [], revision_id=b'revid2a')
        self.builder.build_snapshot([b'revid1'], [], revision_id=b'revid2b')
        self.builder.build_snapshot([b'revid2a', b'revid2b'], [],
                                    revision_id=b'revid3')
        rev_set = [b'revid3', b'revid2a']
        self.assertParentIds([b'revid1', b'revid2b'], rev_set)


class TestWithBrokenRepo(TestCaseWithTransport):
    """These tests seem to be more appropriate as interface tests?"""

    def make_broken_repository(self):
        # XXX: This function is borrowed from Aaron's "Reconcile can fix bad
        # parent references" branch which is due to land in bzr.dev soon.  Once
        # it does, this duplication should be removed.
        repo = self.make_repository('broken-repo')
        cleanups = []
        try:
            repo.lock_write()
            cleanups.append(repo.unlock)
            repo.start_write_group()
            cleanups.append(repo.commit_write_group)
            # make rev1a: A well-formed revision, containing 'file1'
            inv = inventory.Inventory(revision_id=b'rev1a')
            inv.root.revision = b'rev1a'
            self.add_file(repo, inv, 'file1', b'rev1a', [])
            repo.texts.add_lines((inv.root.file_id, b'rev1a'), [], [])
            repo.add_inventory(b'rev1a', inv, [])
            revision = _mod_revision.Revision(
                b'rev1a',
                committer='jrandom@example.com', timestamp=0,
                inventory_sha1='', timezone=0, message='foo', parent_ids=[])
            repo.add_revision(b'rev1a', revision, inv)

            # make rev1b, which has no Revision, but has an Inventory, and
            # file1
            inv = inventory.Inventory(revision_id=b'rev1b')
            inv.root.revision = b'rev1b'
            self.add_file(repo, inv, 'file1', b'rev1b', [])
            repo.add_inventory(b'rev1b', inv, [])

            # make rev2, with file1 and file2
            # file2 is sane
            # file1 has 'rev1b' as an ancestor, even though this is not
            # mentioned by 'rev1a', making it an unreferenced ancestor
            inv = inventory.Inventory()
            self.add_file(repo, inv, 'file1', b'rev2', [b'rev1a', b'rev1b'])
            self.add_file(repo, inv, 'file2', b'rev2', [])
            self.add_revision(repo, b'rev2', inv, [b'rev1a'])

            # make ghost revision rev1c
            inv = inventory.Inventory()
            self.add_file(repo, inv, 'file2', b'rev1c', [])

            # make rev3 with file2
            # file2 refers to 'rev1c', which is a ghost in this repository, so
            # file2 cannot have rev1c as its ancestor.
            inv = inventory.Inventory()
            self.add_file(repo, inv, 'file2', b'rev3', [b'rev1c'])
            self.add_revision(repo, b'rev3', inv, [b'rev1c'])
            return repo
        finally:
            for cleanup in reversed(cleanups):
                cleanup()

    def add_revision(self, repo, revision_id, inv, parent_ids):
        inv.revision_id = revision_id
        inv.root.revision = revision_id
        repo.texts.add_lines((inv.root.file_id, revision_id), [], [])
        repo.add_inventory(revision_id, inv, parent_ids)
        revision = _mod_revision.Revision(
            revision_id,
            committer='jrandom@example.com', timestamp=0, inventory_sha1='',
            timezone=0, message='foo', parent_ids=parent_ids)
        repo.add_revision(revision_id, revision, inv)

    def add_file(self, repo, inv, filename, revision, parents):
        file_id = filename.encode('utf-8') + b'-id'
        content = [b'line\n']
        entry = inventory.InventoryFile(file_id, filename, b'TREE_ROOT')
        entry.revision = revision
        entry.text_sha1 = osutils.sha_strings(content)
        entry.text_size = 0
        inv.add(entry)
        text_key = (file_id, revision)
        parent_keys = [(file_id, parent) for parent in parents]
        repo.texts.add_lines(text_key, parent_keys, content)

    def test_insert_from_broken_repo(self):
        """Inserting a data stream from a broken repository won't silently
        corrupt the target repository.
        """
        broken_repo = self.make_broken_repository()
        empty_repo = self.make_repository('empty-repo')
        try:
            empty_repo.fetch(broken_repo)
        except (errors.RevisionNotPresent, errors.BzrCheckError):
            # Test successful: compression parent not being copied leads to
            # error.
            return
        empty_repo.lock_read()
        self.addCleanup(empty_repo.unlock)
        text = next(empty_repo.texts.get_record_stream(
            [(b'file2-id', b'rev3')], 'topological', True))
        self.assertEqual(b'line\n', text.get_bytes_as('fulltext'))


class TestRepositoryPackCollection(TestCaseWithTransport):

    def get_format(self):
        return controldir.format_registry.make_controldir('pack-0.92')

    def get_packs(self):
        format = self.get_format()
        repo = self.make_repository('.', format=format)
        return repo._pack_collection

    def make_packs_and_alt_repo(self, write_lock=False):
        """Create a pack repo with 3 packs, and access it via a second repo."""
        tree = self.make_branch_and_tree('.', format=self.get_format())
        tree.lock_write()
        self.addCleanup(tree.unlock)
        rev1 = tree.commit('one')
        rev2 = tree.commit('two')
        rev3 = tree.commit('three')
        r = repository.Repository.open('.')
        if write_lock:
            r.lock_write()
        else:
            r.lock_read()
        self.addCleanup(r.unlock)
        packs = r._pack_collection
        packs.ensure_loaded()
        return tree, r, packs, [rev1, rev2, rev3]

    def test__clear_obsolete_packs(self):
        packs = self.get_packs()
        obsolete_pack_trans = packs.transport.clone('obsolete_packs')
        obsolete_pack_trans.put_bytes('a-pack.pack', b'content\n')
        obsolete_pack_trans.put_bytes('a-pack.rix', b'content\n')
        obsolete_pack_trans.put_bytes('a-pack.iix', b'content\n')
        obsolete_pack_trans.put_bytes('another-pack.pack', b'foo\n')
        obsolete_pack_trans.put_bytes('not-a-pack.rix', b'foo\n')
        res = packs._clear_obsolete_packs()
        self.assertEqual(['a-pack', 'another-pack'], sorted(res))
        self.assertEqual([], obsolete_pack_trans.list_dir('.'))

    def test__clear_obsolete_packs_preserve(self):
        packs = self.get_packs()
        obsolete_pack_trans = packs.transport.clone('obsolete_packs')
        obsolete_pack_trans.put_bytes('a-pack.pack', b'content\n')
        obsolete_pack_trans.put_bytes('a-pack.rix', b'content\n')
        obsolete_pack_trans.put_bytes('a-pack.iix', b'content\n')
        obsolete_pack_trans.put_bytes('another-pack.pack', b'foo\n')
        obsolete_pack_trans.put_bytes('not-a-pack.rix', b'foo\n')
        res = packs._clear_obsolete_packs(preserve={'a-pack'})
        self.assertEqual(['a-pack', 'another-pack'], sorted(res))
        self.assertEqual(['a-pack.iix', 'a-pack.pack', 'a-pack.rix'],
                         sorted(obsolete_pack_trans.list_dir('.')))

    def test__max_pack_count(self):
        """The maximum pack count is a function of the number of revisions."""
        # no revisions - one pack, so that we can have a revision free repo
        # without it blowing up
        packs = self.get_packs()
        self.assertEqual(1, packs._max_pack_count(0))
        # after that the sum of the digits, - check the first 1-9
        self.assertEqual(1, packs._max_pack_count(1))
        self.assertEqual(2, packs._max_pack_count(2))
        self.assertEqual(3, packs._max_pack_count(3))
        self.assertEqual(4, packs._max_pack_count(4))
        self.assertEqual(5, packs._max_pack_count(5))
        self.assertEqual(6, packs._max_pack_count(6))
        self.assertEqual(7, packs._max_pack_count(7))
        self.assertEqual(8, packs._max_pack_count(8))
        self.assertEqual(9, packs._max_pack_count(9))
        # check the boundary cases with two digits for the next decade
        self.assertEqual(1, packs._max_pack_count(10))
        self.assertEqual(2, packs._max_pack_count(11))
        self.assertEqual(10, packs._max_pack_count(19))
        self.assertEqual(2, packs._max_pack_count(20))
        self.assertEqual(3, packs._max_pack_count(21))
        # check some arbitrary big numbers
        self.assertEqual(25, packs._max_pack_count(112894))

    def test_repr(self):
        packs = self.get_packs()
        self.assertContainsRe(repr(packs),
                              'RepositoryPackCollection(.*Repository(.*))')

    def test__obsolete_packs(self):
        tree, r, packs, revs = self.make_packs_and_alt_repo(write_lock=True)
        names = packs.names()
        pack = packs.get_pack_by_name(names[0])
        # Schedule this one for removal
        packs._remove_pack_from_memory(pack)
        # Simulate a concurrent update by renaming the .pack file and one of
        # the indices
        packs.transport.rename('packs/{}.pack'.format(names[0]),
                               'obsolete_packs/{}.pack'.format(names[0]))
        packs.transport.rename('indices/{}.iix'.format(names[0]),
                               'obsolete_packs/{}.iix'.format(names[0]))
        # Now trigger the obsoletion, and ensure that all the remaining files
        # are still renamed
        packs._obsolete_packs([pack])
        self.assertEqual([n + '.pack' for n in names[1:]],
                         sorted(packs._pack_transport.list_dir('.')))
        # names[0] should not be present in the index anymore
        self.assertEqual(names[1:],
                         sorted({osutils.splitext(n)[0] for n in
                                 packs._index_transport.list_dir('.')}))

    def test__obsolete_packs_missing_directory(self):
        tree, r, packs, revs = self.make_packs_and_alt_repo(write_lock=True)
        r.control_transport.rmdir('obsolete_packs')
        names = packs.names()
        pack = packs.get_pack_by_name(names[0])
        # Schedule this one for removal
        packs._remove_pack_from_memory(pack)
        # Now trigger the obsoletion, and ensure that all the remaining files
        # are still renamed
        packs._obsolete_packs([pack])
        self.assertEqual([n + '.pack' for n in names[1:]],
                         sorted(packs._pack_transport.list_dir('.')))
        # names[0] should not be present in the index anymore
        self.assertEqual(names[1:],
                         sorted({osutils.splitext(n)[0] for n in
                                 packs._index_transport.list_dir('.')}))

    def test_pack_distribution_zero(self):
        packs = self.get_packs()
        self.assertEqual([0], packs.pack_distribution(0))

    def test_ensure_loaded_unlocked(self):
        packs = self.get_packs()
        self.assertRaises(errors.ObjectNotLocked,
                          packs.ensure_loaded)

    def test_pack_distribution_one_to_nine(self):
        packs = self.get_packs()
        self.assertEqual([1],
                         packs.pack_distribution(1))
        self.assertEqual([1, 1],
                         packs.pack_distribution(2))
        self.assertEqual([1, 1, 1],
                         packs.pack_distribution(3))
        self.assertEqual([1, 1, 1, 1],
                         packs.pack_distribution(4))
        self.assertEqual([1, 1, 1, 1, 1],
                         packs.pack_distribution(5))
        self.assertEqual([1, 1, 1, 1, 1, 1],
                         packs.pack_distribution(6))
        self.assertEqual([1, 1, 1, 1, 1, 1, 1],
                         packs.pack_distribution(7))
        self.assertEqual([1, 1, 1, 1, 1, 1, 1, 1],
                         packs.pack_distribution(8))
        self.assertEqual([1, 1, 1, 1, 1, 1, 1, 1, 1],
                         packs.pack_distribution(9))

    def test_pack_distribution_stable_at_boundaries(self):
        """When there are multi-rev packs the counts are stable."""
        packs = self.get_packs()
        # in 10s:
        self.assertEqual([10], packs.pack_distribution(10))
        self.assertEqual([10, 1], packs.pack_distribution(11))
        self.assertEqual([10, 10], packs.pack_distribution(20))
        self.assertEqual([10, 10, 1], packs.pack_distribution(21))
        # 100s
        self.assertEqual([100], packs.pack_distribution(100))
        self.assertEqual([100, 1], packs.pack_distribution(101))
        self.assertEqual([100, 10, 1], packs.pack_distribution(111))
        self.assertEqual([100, 100], packs.pack_distribution(200))
        self.assertEqual([100, 100, 1], packs.pack_distribution(201))
        self.assertEqual([100, 100, 10, 1], packs.pack_distribution(211))

    def test_plan_pack_operations_2009_revisions_skip_all_packs(self):
        packs = self.get_packs()
        existing_packs = [(2000, "big"), (9, "medium")]
        # rev count - 2009 -> 2x1000 + 9x1
        pack_operations = packs.plan_autopack_combinations(
            existing_packs, [1000, 1000, 1, 1, 1, 1, 1, 1, 1, 1, 1])
        self.assertEqual([], pack_operations)

    def test_plan_pack_operations_2010_revisions_skip_all_packs(self):
        packs = self.get_packs()
        existing_packs = [(2000, "big"), (9, "medium"), (1, "single")]
        # rev count - 2010 -> 2x1000 + 1x10
        pack_operations = packs.plan_autopack_combinations(
            existing_packs, [1000, 1000, 10])
        self.assertEqual([], pack_operations)

    def test_plan_pack_operations_2010_combines_smallest_two(self):
        packs = self.get_packs()
        existing_packs = [(1999, "big"), (9, "medium"), (1, "single2"),
                          (1, "single1")]
        # rev count - 2010 -> 2x1000 + 1x10 (3)
        pack_operations = packs.plan_autopack_combinations(
            existing_packs, [1000, 1000, 10])
        self.assertEqual([[2, ["single2", "single1"]]], pack_operations)

    def test_plan_pack_operations_creates_a_single_op(self):
        packs = self.get_packs()
        existing_packs = [(50, 'a'), (40, 'b'), (30, 'c'), (10, 'd'),
                          (10, 'e'), (6, 'f'), (4, 'g')]
        # rev count 150 -> 1x100 and 5x10
        # The two size 10 packs do not need to be touched. The 50, 40, 30 would
        # be combined into a single 120 size pack, and the 6 & 4 would
        # becombined into a size 10 pack. However, if we have to rewrite them,
        # we save a pack file with no increased I/O by putting them into the
        # same file.
        distribution = packs.pack_distribution(150)
        pack_operations = packs.plan_autopack_combinations(existing_packs,
                                                           distribution)
        self.assertEqual([[130, ['a', 'b', 'c', 'f', 'g']]], pack_operations)

    def test_all_packs_none(self):
        format = self.get_format()
        tree = self.make_branch_and_tree('.', format=format)
        tree.lock_read()
        self.addCleanup(tree.unlock)
        packs = tree.branch.repository._pack_collection
        packs.ensure_loaded()
        self.assertEqual([], packs.all_packs())

    def test_all_packs_one(self):
        format = self.get_format()
        tree = self.make_branch_and_tree('.', format=format)
        tree.commit('start')
        tree.lock_read()
        self.addCleanup(tree.unlock)
        packs = tree.branch.repository._pack_collection
        packs.ensure_loaded()
        self.assertEqual([
            packs.get_pack_by_name(packs.names()[0])],
            packs.all_packs())

    def test_all_packs_two(self):
        format = self.get_format()
        tree = self.make_branch_and_tree('.', format=format)
        tree.commit('start')
        tree.commit('continue')
        tree.lock_read()
        self.addCleanup(tree.unlock)
        packs = tree.branch.repository._pack_collection
        packs.ensure_loaded()
        self.assertEqual([
            packs.get_pack_by_name(packs.names()[0]),
            packs.get_pack_by_name(packs.names()[1]),
            ], packs.all_packs())

    def test_get_pack_by_name(self):
        format = self.get_format()
        tree = self.make_branch_and_tree('.', format=format)
        tree.commit('start')
        tree.lock_read()
        self.addCleanup(tree.unlock)
        packs = tree.branch.repository._pack_collection
        packs.reset()
        packs.ensure_loaded()
        name = packs.names()[0]
        pack_1 = packs.get_pack_by_name(name)
        # the pack should be correctly initialised
        sizes = packs._names[name]
        rev_index = GraphIndex(packs._index_transport, name + '.rix', sizes[0])
        inv_index = GraphIndex(packs._index_transport, name + '.iix', sizes[1])
        txt_index = GraphIndex(packs._index_transport, name + '.tix', sizes[2])
        sig_index = GraphIndex(packs._index_transport, name + '.six', sizes[3])
        self.assertEqual(
            pack_repo.ExistingPack(
                packs._pack_transport, name, rev_index, inv_index, txt_index,
                sig_index), pack_1)
        # and the same instance should be returned on successive calls.
        self.assertTrue(pack_1 is packs.get_pack_by_name(name))

    def test_reload_pack_names_new_entry(self):
        tree, r, packs, revs = self.make_packs_and_alt_repo()
        names = packs.names()
        # Add a new pack file into the repository
        rev4 = tree.commit('four')
        new_names = tree.branch.repository._pack_collection.names()
        new_name = set(new_names).difference(names)
        self.assertEqual(1, len(new_name))
        new_name = new_name.pop()
        # The old collection hasn't noticed yet
        self.assertEqual(names, packs.names())
        self.assertTrue(packs.reload_pack_names())
        self.assertEqual(new_names, packs.names())
        # And the repository can access the new revision
        self.assertEqual({rev4: (revs[-1],)}, r.get_parent_map([rev4]))
        self.assertFalse(packs.reload_pack_names())

    def test_reload_pack_names_added_and_removed(self):
        tree, r, packs, revs = self.make_packs_and_alt_repo()
        names = packs.names()
        # Now repack the whole thing
        tree.branch.repository.pack()
        new_names = tree.branch.repository._pack_collection.names()
        # The other collection hasn't noticed yet
        self.assertEqual(names, packs.names())
        self.assertTrue(packs.reload_pack_names())
        self.assertEqual(new_names, packs.names())
        self.assertEqual({revs[-1]: (revs[-2],)}, r.get_parent_map([revs[-1]]))
        self.assertFalse(packs.reload_pack_names())

    def test_reload_pack_names_preserves_pending(self):
        # TODO: Update this to also test for pending-deleted names
        tree, r, packs, revs = self.make_packs_and_alt_repo(write_lock=True)
        # We will add one pack (via start_write_group + insert_record_stream),
        # and remove another pack (via _remove_pack_from_memory)
        orig_names = packs.names()
        orig_at_load = packs._packs_at_load
        to_remove_name = next(iter(orig_names))
        r.start_write_group()
        self.addCleanup(r.abort_write_group)
        r.texts.insert_record_stream([versionedfile.FulltextContentFactory(
            (b'text', b'rev'), (), None, b'content\n')])
        new_pack = packs._new_pack
        self.assertTrue(new_pack.data_inserted())
        new_pack.finish()
        packs.allocate(new_pack)
        packs._new_pack = None
        removed_pack = packs.get_pack_by_name(to_remove_name)
        packs._remove_pack_from_memory(removed_pack)
        names = packs.names()
        all_nodes, deleted_nodes, new_nodes, _ = packs._diff_pack_names()
        new_names = {x[0] for x in new_nodes}
        self.assertEqual(names, sorted([x[0] for x in all_nodes]))
        self.assertEqual(set(names) - set(orig_names), new_names)
        self.assertEqual({new_pack.name}, new_names)
        self.assertEqual([to_remove_name],
                         sorted([x[0] for x in deleted_nodes]))
        packs.reload_pack_names()
        reloaded_names = packs.names()
        self.assertEqual(orig_at_load, packs._packs_at_load)
        self.assertEqual(names, reloaded_names)
        all_nodes, deleted_nodes, new_nodes, _ = packs._diff_pack_names()
        new_names = {x[0] for x in new_nodes}
        self.assertEqual(names, sorted([x[0] for x in all_nodes]))
        self.assertEqual(set(names) - set(orig_names), new_names)
        self.assertEqual({new_pack.name}, new_names)
        self.assertEqual([to_remove_name],
                         sorted([x[0] for x in deleted_nodes]))

    def test_autopack_obsoletes_new_pack(self):
        tree, r, packs, revs = self.make_packs_and_alt_repo(write_lock=True)
        packs._max_pack_count = lambda x: 1
        packs.pack_distribution = lambda x: [10]
        r.start_write_group()
        r.revisions.insert_record_stream([versionedfile.FulltextContentFactory(
            (b'bogus-rev',), (), None, b'bogus-content\n')])
        # This should trigger an autopack, which will combine everything into a
        # single pack file.
        r.commit_write_group()
        names = packs.names()
        self.assertEqual(1, len(names))
        self.assertEqual([names[0] + '.pack'],
                         packs._pack_transport.list_dir('.'))

    def test_autopack_reloads_and_stops(self):
        tree, r, packs, revs = self.make_packs_and_alt_repo(write_lock=True)
        # After we have determined what needs to be autopacked, trigger a
        # full-pack via the other repo which will cause us to re-evaluate and
        # decide we don't need to do anything
        orig_execute = packs._execute_pack_operations

        def _munged_execute_pack_ops(*args, **kwargs):
            tree.branch.repository.pack()
            return orig_execute(*args, **kwargs)
        packs._execute_pack_operations = _munged_execute_pack_ops
        packs._max_pack_count = lambda x: 1
        packs.pack_distribution = lambda x: [10]
        self.assertFalse(packs.autopack())
        self.assertEqual(1, len(packs.names()))
        self.assertEqual(tree.branch.repository._pack_collection.names(),
                         packs.names())

    def test__save_pack_names(self):
        tree, r, packs, revs = self.make_packs_and_alt_repo(write_lock=True)
        names = packs.names()
        pack = packs.get_pack_by_name(names[0])
        packs._remove_pack_from_memory(pack)
        packs._save_pack_names(obsolete_packs=[pack])
        cur_packs = packs._pack_transport.list_dir('.')
        self.assertEqual([n + '.pack' for n in names[1:]], sorted(cur_packs))
        # obsolete_packs will also have stuff like .rix and .iix present.
        obsolete_packs = packs.transport.list_dir('obsolete_packs')
        obsolete_names = {osutils.splitext(n)[0] for n in obsolete_packs}
        self.assertEqual([pack.name], sorted(obsolete_names))

    def test__save_pack_names_already_obsoleted(self):
        tree, r, packs, revs = self.make_packs_and_alt_repo(write_lock=True)
        names = packs.names()
        pack = packs.get_pack_by_name(names[0])
        packs._remove_pack_from_memory(pack)
        # We are going to simulate a concurrent autopack by manually obsoleting
        # the pack directly.
        packs._obsolete_packs([pack])
        packs._save_pack_names(clear_obsolete_packs=True,
                               obsolete_packs=[pack])
        cur_packs = packs._pack_transport.list_dir('.')
        self.assertEqual([n + '.pack' for n in names[1:]], sorted(cur_packs))
        # Note that while we set clear_obsolete_packs=True, it should not
        # delete a pack file that we have also scheduled for obsoletion.
        obsolete_packs = packs.transport.list_dir('obsolete_packs')
        obsolete_names = {osutils.splitext(n)[0] for n in obsolete_packs}
        self.assertEqual([pack.name], sorted(obsolete_names))

    def test_pack_no_obsolete_packs_directory(self):
        """Bug #314314, don't fail if obsolete_packs directory does
        not exist."""
        tree, r, packs, revs = self.make_packs_and_alt_repo(write_lock=True)
        r.control_transport.rmdir('obsolete_packs')
        packs._clear_obsolete_packs()


class TestPack(TestCaseWithTransport):
    """Tests for the Pack object."""

    def assertCurrentlyEqual(self, left, right):
        self.assertTrue(left == right)
        self.assertTrue(right == left)
        self.assertFalse(left != right)
        self.assertFalse(right != left)

    def assertCurrentlyNotEqual(self, left, right):
        self.assertFalse(left == right)
        self.assertFalse(right == left)
        self.assertTrue(left != right)
        self.assertTrue(right != left)

    def test___eq____ne__(self):
        left = pack_repo.ExistingPack('', '', '', '', '', '')
        right = pack_repo.ExistingPack('', '', '', '', '', '')
        self.assertCurrentlyEqual(left, right)
        # change all attributes and ensure equality changes as we do.
        left.revision_index = 'a'
        self.assertCurrentlyNotEqual(left, right)
        right.revision_index = 'a'
        self.assertCurrentlyEqual(left, right)
        left.inventory_index = 'a'
        self.assertCurrentlyNotEqual(left, right)
        right.inventory_index = 'a'
        self.assertCurrentlyEqual(left, right)
        left.text_index = 'a'
        self.assertCurrentlyNotEqual(left, right)
        right.text_index = 'a'
        self.assertCurrentlyEqual(left, right)
        left.signature_index = 'a'
        self.assertCurrentlyNotEqual(left, right)
        right.signature_index = 'a'
        self.assertCurrentlyEqual(left, right)
        left.name = 'a'
        self.assertCurrentlyNotEqual(left, right)
        right.name = 'a'
        self.assertCurrentlyEqual(left, right)
        left.transport = 'a'
        self.assertCurrentlyNotEqual(left, right)
        right.transport = 'a'
        self.assertCurrentlyEqual(left, right)

    def test_file_name(self):
        pack = pack_repo.ExistingPack('', 'a_name', '', '', '', '')
        self.assertEqual('a_name.pack', pack.file_name())


class TestNewPack(TestCaseWithTransport):
    """Tests for pack_repo.NewPack."""

    def test_new_instance_attributes(self):
        upload_transport = self.get_transport('upload')
        pack_transport = self.get_transport('pack')
        index_transport = self.get_transport('index')
        upload_transport.mkdir('.')
        collection = pack_repo.RepositoryPackCollection(
            repo=None,
            transport=self.get_transport('.'),
            index_transport=index_transport,
            upload_transport=upload_transport,
            pack_transport=pack_transport,
            index_builder_class=BTreeBuilder,
            index_class=BTreeGraphIndex,
            use_chk_index=False)
        pack = pack_repo.NewPack(collection)
        self.addCleanup(pack.abort)  # Make sure the write stream gets closed
        self.assertIsInstance(pack.revision_index, BTreeBuilder)
        self.assertIsInstance(pack.inventory_index, BTreeBuilder)
        self.assertIsInstance(pack._hash, type(osutils.md5()))
        self.assertTrue(pack.upload_transport is upload_transport)
        self.assertTrue(pack.index_transport is index_transport)
        self.assertTrue(pack.pack_transport is pack_transport)
        self.assertEqual(None, pack.index_sizes)
        self.assertEqual(20, len(pack.random_name))
        self.assertIsInstance(pack.random_name, str)
        self.assertIsInstance(pack.start_time, float)


class TestPacker(TestCaseWithTransport):
    """Tests for the packs repository Packer class."""

    def test_pack_optimizes_pack_order(self):
        builder = self.make_branch_builder('.', format="1.9")
        builder.start_series()
        builder.build_snapshot(None, [
            ('add', ('', b'root-id', 'directory', None)),
            ('add', ('f', b'f-id', 'file', b'content\n'))],
            revision_id=b'A')
        builder.build_snapshot([b'A'],
                               [('modify', ('f', b'new-content\n'))],
                               revision_id=b'B')
        builder.build_snapshot([b'B'],
                               [('modify', ('f', b'third-content\n'))],
                               revision_id=b'C')
        builder.build_snapshot([b'C'],
                               [('modify', ('f', b'fourth-content\n'))],
                               revision_id=b'D')
        b = builder.get_branch()
        b.lock_read()
        builder.finish_series()
        self.addCleanup(b.unlock)
        # At this point, we should have 4 pack files available
        # Because of how they were built, they correspond to
        # ['D', 'C', 'B', 'A']
        packs = b.repository._pack_collection.packs
        packer = knitpack_repo.KnitPacker(b.repository._pack_collection,
                                          packs, 'testing',
                                          revision_ids=[b'B', b'C'])
        # Now, when we are copying the B & C revisions, their pack files should
        # be moved to the front of the stack
        # The new ordering moves B & C to the front of the .packs attribute,
        # and leaves the others in the original order.
        new_packs = [packs[1], packs[2], packs[0], packs[3]]
        packer.pack()
        self.assertEqual(new_packs, packer.packs)


class TestOptimisingPacker(TestCaseWithTransport):
    """Tests for the OptimisingPacker class."""

    def get_pack_collection(self):
        repo = self.make_repository('.')
        return repo._pack_collection

    def test_open_pack_will_optimise(self):
        packer = knitpack_repo.OptimisingKnitPacker(self.get_pack_collection(),
                                                    [], '.test')
        new_pack = packer.open_pack()
        self.addCleanup(new_pack.abort)  # ensure cleanup
        self.assertIsInstance(new_pack, pack_repo.NewPack)
        self.assertTrue(new_pack.revision_index._optimize_for_size)
        self.assertTrue(new_pack.inventory_index._optimize_for_size)
        self.assertTrue(new_pack.text_index._optimize_for_size)
        self.assertTrue(new_pack.signature_index._optimize_for_size)


class TestGCCHKPacker(TestCaseWithTransport):

    def make_abc_branch(self):
        builder = self.make_branch_builder('source')
        builder.start_series()
        builder.build_snapshot(None, [
            ('add', ('', b'root-id', 'directory', None)),
            ('add', ('file', b'file-id', 'file', b'content\n')),
            ], revision_id=b'A')
        builder.build_snapshot([b'A'], [
            ('add', ('dir', b'dir-id', 'directory', None))],
            revision_id=b'B')
        builder.build_snapshot([b'B'], [
            ('modify', ('file', b'new content\n'))],
            revision_id=b'C')
        builder.finish_series()
        return builder.get_branch()

    def make_branch_with_disjoint_inventory_and_revision(self):
        """a repo with separate packs for a revisions Revision and Inventory.

        There will be one pack file that holds the Revision content, and one
        for the Inventory content.

        :return: (repository,
                  pack_name_with_rev_A_Revision,
                  pack_name_with_rev_A_Inventory,
                  pack_name_with_rev_C_content)
        """
        b_source = self.make_abc_branch()
        b_base = b_source.controldir.sprout(
            'base', revision_id=b'A').open_branch()
        b_stacked = b_base.controldir.sprout(
            'stacked', stacked=True).open_branch()
        b_stacked.lock_write()
        self.addCleanup(b_stacked.unlock)
        b_stacked.fetch(b_source, b'B')
        # Now re-open the stacked repo directly (no fallbacks) so that we can
        # fill in the A rev.
        repo_not_stacked = b_stacked.controldir.open_repository()
        repo_not_stacked.lock_write()
        self.addCleanup(repo_not_stacked.unlock)
        # Now we should have a pack file with A's inventory, but not its
        # Revision
        self.assertEqual([(b'A',), (b'B',)],
                         sorted(repo_not_stacked.inventories.keys()))
        self.assertEqual([(b'B',)],
                         sorted(repo_not_stacked.revisions.keys()))
        stacked_pack_names = repo_not_stacked._pack_collection.names()
        # We have a couple names here, figure out which has A's inventory
        for name in stacked_pack_names:
            pack = repo_not_stacked._pack_collection.get_pack_by_name(name)
            keys = [n[1] for n in pack.inventory_index.iter_all_entries()]
            if (b'A',) in keys:
                inv_a_pack_name = name
                break
        else:
            self.fail('Could not find pack containing A\'s inventory')
        repo_not_stacked.fetch(b_source.repository, b'A')
        self.assertEqual([(b'A',), (b'B',)],
                         sorted(repo_not_stacked.revisions.keys()))
        new_pack_names = set(repo_not_stacked._pack_collection.names())
        rev_a_pack_names = new_pack_names.difference(stacked_pack_names)
        self.assertEqual(1, len(rev_a_pack_names))
        rev_a_pack_name = list(rev_a_pack_names)[0]
        # Now fetch 'C', so we have a couple pack files to join
        repo_not_stacked.fetch(b_source.repository, b'C')
        rev_c_pack_names = set(repo_not_stacked._pack_collection.names())
        rev_c_pack_names = rev_c_pack_names.difference(new_pack_names)
        self.assertEqual(1, len(rev_c_pack_names))
        rev_c_pack_name = list(rev_c_pack_names)[0]
        return (repo_not_stacked, rev_a_pack_name, inv_a_pack_name,
                rev_c_pack_name)

    def test_pack_with_distant_inventories(self):
        # See https://bugs.launchpad.net/bzr/+bug/437003
        # When repacking, it is possible to have an inventory in a different
        # pack file than the associated revision. An autopack can then come
        # along, and miss that inventory, and complain.
        (repo, rev_a_pack_name, inv_a_pack_name, rev_c_pack_name
         ) = self.make_branch_with_disjoint_inventory_and_revision()
        a_pack = repo._pack_collection.get_pack_by_name(rev_a_pack_name)
        c_pack = repo._pack_collection.get_pack_by_name(rev_c_pack_name)
        packer = groupcompress_repo.GCCHKPacker(repo._pack_collection,
                                                [a_pack, c_pack], '.test-pack')
        # This would raise ValueError in bug #437003, but should not raise an
        # error once fixed.
        packer.pack()

    def test_pack_with_missing_inventory(self):
        # Similar to test_pack_with_missing_inventory, but this time, we force
        # the A inventory to actually be gone from the repository.
        (repo, rev_a_pack_name, inv_a_pack_name, rev_c_pack_name
         ) = self.make_branch_with_disjoint_inventory_and_revision()
        inv_a_pack = repo._pack_collection.get_pack_by_name(inv_a_pack_name)
        repo._pack_collection._remove_pack_from_memory(inv_a_pack)
        packer = groupcompress_repo.GCCHKPacker(repo._pack_collection,
                                                repo._pack_collection.all_packs(), '.test-pack')
        e = self.assertRaises(ValueError, packer.pack)
        packer.new_pack.abort()
        self.assertContainsRe(str(e),
                              r"We are missing inventories for revisions: .*'A'")


class TestCrossFormatPacks(TestCaseWithTransport):

    def log_pack(self, hint=None):
        self.calls.append(('pack', hint))
        self.orig_pack(hint=hint)
        if self.expect_hint:
            self.assertTrue(hint)

    def run_stream(self, src_fmt, target_fmt, expect_pack_called):
        self.expect_hint = expect_pack_called
        self.calls = []
        source_tree = self.make_branch_and_tree('src', format=src_fmt)
        source_tree.lock_write()
        self.addCleanup(source_tree.unlock)
        tip = source_tree.commit('foo')
        target = self.make_repository('target', format=target_fmt)
        target.lock_write()
        self.addCleanup(target.unlock)
        source = source_tree.branch.repository._get_source(target._format)
        self.orig_pack = target.pack
        self.overrideAttr(target, "pack", self.log_pack)
        search = target.search_missing_revision_ids(
            source_tree.branch.repository, revision_ids=[tip])
        stream = source.get_stream(search)
        from_format = source_tree.branch.repository._format
        sink = target._get_sink()
        sink.insert_stream(stream, from_format, [])
        if expect_pack_called:
            self.assertLength(1, self.calls)
        else:
            self.assertLength(0, self.calls)

    def run_fetch(self, src_fmt, target_fmt, expect_pack_called):
        self.expect_hint = expect_pack_called
        self.calls = []
        source_tree = self.make_branch_and_tree('src', format=src_fmt)
        source_tree.lock_write()
        self.addCleanup(source_tree.unlock)
        source_tree.commit('foo')
        target = self.make_repository('target', format=target_fmt)
        target.lock_write()
        self.addCleanup(target.unlock)
        source = source_tree.branch.repository
        self.orig_pack = target.pack
        self.overrideAttr(target, "pack", self.log_pack)
        target.fetch(source)
        if expect_pack_called:
            self.assertLength(1, self.calls)
        else:
            self.assertLength(0, self.calls)

    def test_sink_format_hint_no(self):
        # When the target format says packing makes no difference, pack is not
        # called.
        self.run_stream('1.9', 'rich-root-pack', False)

    def test_sink_format_hint_yes(self):
        # When the target format says packing makes a difference, pack is
        # called.
        self.run_stream('1.9', '2a', True)

    def test_sink_format_same_no(self):
        # When the formats are the same, pack is not called.
        self.run_stream('2a', '2a', False)

    def test_IDS_format_hint_no(self):
        # When the target format says packing makes no difference, pack is not
        # called.
        self.run_fetch('1.9', 'rich-root-pack', False)

    def test_IDS_format_hint_yes(self):
        # When the target format says packing makes a difference, pack is
        # called.
        self.run_fetch('1.9', '2a', True)

    def test_IDS_format_same_no(self):
        # When the formats are the same, pack is not called.
        self.run_fetch('2a', '2a', False)


class Test_LazyListJoin(tests.TestCase):

    def test__repr__(self):
        lazy = repository._LazyListJoin(['a'], ['b'])
        self.assertEqual("breezy.repository._LazyListJoin((['a'], ['b']))",
                         repr(lazy))


class TestFeatures(tests.TestCaseWithTransport):

    def test_open_with_present_feature(self):
        self.addCleanup(
            bzrrepository.RepositoryFormatMetaDir.unregister_feature,
            b"makes-cheese-sandwich")
        bzrrepository.RepositoryFormatMetaDir.register_feature(
            b"makes-cheese-sandwich")
        repo = self.make_repository('.')
        repo.lock_write()
        repo._format.features[b"makes-cheese-sandwich"] = b"required"
        repo._format.check_support_status(False)
        repo.unlock()

    def test_open_with_missing_required_feature(self):
        repo = self.make_repository('.')
        repo.lock_write()
        repo._format.features[b"makes-cheese-sandwich"] = b"required"
        self.assertRaises(bzrdir.MissingFeature,
                          repo._format.check_support_status, False)



# Copyright (C) 2006-2011 Canonical Ltd
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

import codecs
import io

from .. import tests
from ..progress import ProgressTask
from ..ui.text import TextProgressView
from . import ui_testing


class TestTextProgressView(tests.TestCase):
    """Tests for text display of progress bars.

    These try to exercise the progressview independently of its construction,
    which is arranged by the TextUIFactory.
    """
    # The ProgressTask now connects directly to the ProgressView, so we can
    # check them independently of the factory or of the determination of what
    # view to use.

    def make_view_only(self, out, width=79):
        view = TextProgressView(out)
        view._avail_width = lambda: width
        return view

    def make_view(self):
        out = ui_testing.StringIOWithEncoding()
        return out, self.make_view_only(out)

    def make_task(self, parent_task, view, msg, curr, total):
        # would normally be done by UIFactory; is done here so that we don't
        # have to have one.
        task = ProgressTask(parent_task, progress_view=view)
        task.msg = msg
        task.current_cnt = curr
        task.total_cnt = total
        return task

    def test_clear(self):
        # <https://bugs.launchpad.net/bzr/+bug/611127> clear must actually
        # send spaces to clear the line
        out, view = self.make_view()
        task = self.make_task(None, view, 'reticulating splines', 5, 20)
        view.show_progress(task)
        self.assertEqual(
            '\r/ reticulating splines 5/20                                                    \r', out.getvalue())
        view.clear()
        self.assertEqual(
            '\r/ reticulating splines 5/20                                                    \r' +
            '\r' + 79 * ' ' + '\r',
            out.getvalue())

    def test_render_progress_no_bar(self):
        """The default view now has a spinner but no bar."""
        out, view = self.make_view()
        # view.enable_bar = False
        task = self.make_task(None, view, 'reticulating splines', 5, 20)
        view.show_progress(task)
        self.assertEqual(
            '\r/ reticulating splines 5/20                                                    \r', out.getvalue())

    def test_render_progress_easy(self):
        """Just one task and one quarter done"""
        out, view = self.make_view()
        view.enable_bar = True
        task = self.make_task(None, view, 'reticulating splines', 5, 20)
        view.show_progress(task)
        self.assertEqual(
            '\r[####/               ] reticulating splines 5/20                               \r', out.getvalue())

    def test_render_progress_nested(self):
        """Tasks proportionally contribute to overall progress"""
        out, view = self.make_view()
        task = self.make_task(None, view, 'reticulating splines', 0, 2)
        task2 = self.make_task(task, view, 'stage2', 1, 2)
        view.show_progress(task2)
        view.enable_bar = True
        # so we're in the first half of the main task, and half way through
        # that
        self.assertEqual(
            '[####-               ] reticulating splines:stage2 1/2                         ', view._render_line())
        # if the nested task is complete, then we're all the way through the
        # first half of the overall work
        task2.update('stage2', 2, 2)
        self.assertEqual(
            '[#########\\          ] reticulating splines:stage2 2/2                         ', view._render_line())

    def test_render_progress_sub_nested(self):
        """Intermediate tasks don't mess up calculation."""
        out, view = self.make_view()
        view.enable_bar = True
        task_a = ProgressTask(None, progress_view=view)
        task_a.update('a', 0, 2)
        task_b = ProgressTask(task_a, progress_view=view)
        task_b.update('b')
        task_c = ProgressTask(task_b, progress_view=view)
        task_c.update('c', 1, 2)
        # the top-level task is in its first half; the middle one has no
        # progress indication, just a label; and the bottom one is half done,
        # so the overall fraction is 1/4
        self.assertEqual(
            '[####|               ] a:b:c 1/2                                               ', view._render_line())

    def test_render_truncated(self):
        # when the bar is too long for the terminal, we prefer not to truncate
        # the counters because they might be interesting, and because
        # truncating the numbers might be misleading
        out, view = self.make_view()
        task_a = ProgressTask(None, progress_view=view)
        task_a.update('start_' + 'a' * 200 + '_end', 2000, 5000)
        line = view._render_line()
        self.assertEqual(
            '- start_aaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaa.. 2000/5000',
            line)
        self.assertEqual(len(line), 79)

    def test_render_with_activity(self):
        # if the progress view has activity, it's shown before the spinner
        out, view = self.make_view()
        task_a = ProgressTask(None, progress_view=view)
        view._last_transport_msg = '   123kB   100kB/s '
        line = view._render_line()
        self.assertEqual(
            '   123kB   100kB/s /                                                           ',
            line)
        self.assertEqual(len(line), 79)

        task_a.update('start_' + 'a' * 200 + '_end', 2000, 5000)
        view._last_transport_msg = '   123kB   100kB/s '
        line = view._render_line()
        self.assertEqual(
            '   123kB   100kB/s \\ start_aaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaa.. 2000/5000',
            line)
        self.assertEqual(len(line), 79)

    def test_render_progress_unicode_enc_utf8(self):
        out = ui_testing.StringIOWithEncoding()
        out.encoding = "utf-8"
        view = self.make_view_only(out, 20)
        task = self.make_task(None, view, "\xa7", 0, 1)
        view.show_progress(task)
        self.assertEqual('\r/ \xa7 0/1            \r', out.getvalue())

    def test_render_progress_unicode_enc_missing(self):
        out = codecs.getwriter("ascii")(io.BytesIO())
        self.assertRaises(AttributeError, getattr, out, "encoding")
        view = self.make_view_only(out, 20)
        task = self.make_task(None, view, "\xa7", 0, 1)
        view.show_progress(task)
        self.assertEqual(b'\r/ ? 0/1             \r', out.getvalue())

    def test_render_progress_unicode_enc_none(self):
        out = ui_testing.StringIOWithEncoding()
        out.encoding = None
        view = self.make_view_only(out, 20)
        task = self.make_task(None, view, "\xa7", 0, 1)
        view.show_progress(task)
        self.assertEqual('\r/ ? 0/1             \r', out.getvalue())



# Copyright (C) 2005-2011, 2016 Canonical Ltd
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

# Mr. Smoketoomuch: I'm sorry?
# Mr. Bounder: You'd better cut down a little then.
# Mr. Smoketoomuch: Oh, I see! Smoke too much so I'd better cut down a little
#                   then!

"""Black-box tests for bzr.

These check that it behaves properly when it's invoked through the regular
command-line interface. This doesn't actually run a new interpreter but
rather starts again from the run_brz function.
"""


# XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
# Note: Please don't add new tests here, it's too big and bulky.  Instead add
# them into small suites in breezy.tests.blackbox.test_FOO for the particular
# UI command/aspect that is being tested.


import os
import re
import sys

import breezy
from breezy import osutils
from breezy.branch import Branch
from breezy.errors import CommandError
from breezy.tests import TestCaseWithTransport
from breezy.tests.http_utils import TestCaseWithWebserver
from breezy.tests.test_sftp_transport import TestCaseWithSFTPServer
from breezy.workingtree import WorkingTree


class TestCommands(TestCaseWithTransport):

    def test_invalid_commands(self):
        self.run_bzr("pants", retcode=3)
        self.run_bzr("--pants off", retcode=3)
        self.run_bzr("diff --message foo", retcode=3)

    def test_revert(self):
        self.run_bzr('init')

        with open('hello', 'w') as f:
            f.write('foo')
        self.run_bzr('add hello')
        self.run_bzr('commit -m setup hello')

        with open('goodbye', 'w') as f:
            f.write('baz')
        self.run_bzr('add goodbye')
        self.run_bzr('commit -m setup goodbye')

        with open('hello', 'w') as f:
            f.write('bar')
        with open('goodbye', 'w') as f:
            f.write('qux')
        self.run_bzr('revert hello')
        self.check_file_contents('hello', b'foo')
        self.check_file_contents('goodbye', b'qux')
        self.run_bzr('revert')
        self.check_file_contents('goodbye', b'baz')

        os.mkdir('revertdir')
        self.run_bzr('add revertdir')
        self.run_bzr('commit -m f')
        os.rmdir('revertdir')
        self.run_bzr('revert')

        if osutils.supports_symlinks(self.test_dir):
            os.symlink('/unlikely/to/exist', 'symlink')
            self.run_bzr('add symlink')
            self.run_bzr('commit -m f')
            os.unlink('symlink')
            self.run_bzr('revert')
            self.assertPathExists('symlink')
            os.unlink('symlink')
            os.symlink('a-different-path', 'symlink')
            self.run_bzr('revert')
            self.assertEqual('/unlikely/to/exist',
                             os.readlink('symlink'))
        else:
            self.log("skipping revert symlink tests")

        with open('hello', 'w') as f:
            f.write('xyz')
        self.run_bzr('commit -m xyz hello')
        self.run_bzr('revert -r 1 hello')
        self.check_file_contents('hello', b'foo')
        self.run_bzr('revert hello')
        self.check_file_contents('hello', b'xyz')
        os.chdir('revertdir')
        self.run_bzr('revert')
        os.chdir('..')

    def example_branch(test):
        test.run_bzr('init')
        with open('hello', 'w') as f:
            f.write('foo')
        test.run_bzr('add hello')
        test.run_bzr('commit -m setup hello')
        with open('goodbye', 'w') as f:
            f.write('baz')
        test.run_bzr('add goodbye')
        test.run_bzr('commit -m setup goodbye')

    def test_pull_verbose(self):
        """Pull changes from one branch to another and watch the output."""

        os.mkdir('a')
        os.chdir('a')

        self.example_branch()

        os.chdir('..')
        self.run_bzr('branch a b')
        os.chdir('b')
        with open('b', 'wb') as f:
            f.write(b'else\n')
        self.run_bzr('add b')
        self.run_bzr(['commit', '-m', 'added b'])

        os.chdir('../a')
        out = self.run_bzr('pull --verbose ../b')[0]
        self.assertNotEqual(out.find('Added Revisions:'), -1)
        self.assertNotEqual(out.find('message:\n  added b'), -1)
        self.assertNotEqual(out.find('added b'), -1)

        # Check that --overwrite --verbose prints out the removed entries
        self.run_bzr('commit -m foo --unchanged')
        os.chdir('../b')
        self.run_bzr('commit -m baz --unchanged')
        self.run_bzr('pull ../a', retcode=3)
        out = self.run_bzr('pull --overwrite --verbose ../a')[0]

        remove_loc = out.find('Removed Revisions:')
        self.assertNotEqual(remove_loc, -1)
        added_loc = out.find('Added Revisions:')
        self.assertNotEqual(added_loc, -1)

        removed_message = out.find('message:\n  baz')
        self.assertNotEqual(removed_message, -1)
        self.assertTrue(remove_loc < removed_message < added_loc)

        added_message = out.find('message:\n  foo')
        self.assertNotEqual(added_message, -1)
        self.assertTrue(added_loc < added_message)

    def test_locations(self):
        """Using and remembering different locations"""
        os.mkdir('a')
        os.chdir('a')
        self.run_bzr('init')
        self.run_bzr('commit -m unchanged --unchanged')
        self.run_bzr('pull', retcode=3)
        self.run_bzr('merge', retcode=3)
        self.run_bzr('branch . ../b')
        os.chdir('../b')
        self.run_bzr('pull')
        self.run_bzr('branch . ../c')
        self.run_bzr('pull ../c')
        self.run_bzr('merge')
        os.chdir('../a')
        self.run_bzr('pull ../b')
        self.run_bzr('pull')
        self.run_bzr('pull ../c')
        self.run_bzr('branch ../c ../d')
        osutils.rmtree('../c')
        self.run_bzr('pull')
        os.chdir('../b')
        self.run_bzr('pull')
        os.chdir('../d')
        self.run_bzr('pull', retcode=3)
        self.run_bzr('pull ../a --remember')
        self.run_bzr('pull')

    def test_unknown_command(self):
        """Handling of unknown command."""
        out, err = self.run_bzr('fluffy-badger', retcode=3)
        self.assertEqual(out, '')
        err.index('unknown command')

    def create_conflicts(self):
        """Create a conflicted tree"""
        os.mkdir('base')
        os.chdir('base')
        with open('hello', 'wb') as f:
            f.write(b"hi world")
        with open('answer', 'wb') as f:
            f.write(b"42")
        self.run_bzr('init')
        self.run_bzr('add')
        self.run_bzr('commit -m base')
        self.run_bzr('branch . ../other')
        self.run_bzr('branch . ../this')
        os.chdir('../other')
        with open('hello', 'wb') as f:
            f.write(b"Hello.")
        with open('answer', 'wb') as f:
            f.write(b"Is anyone there?")
        self.run_bzr('commit -m other')
        os.chdir('../this')
        with open('hello', 'wb') as f:
            f.write(b"Hello, world")
        self.run_bzr('mv answer question')
        with open('question', 'wb') as f:
            f.write(b"What do you get when you multiply six"
                    b"times nine?")
        self.run_bzr('commit -m this')

    def test_status(self):
        os.mkdir('branch1')
        os.chdir('branch1')
        self.run_bzr('init')
        self.run_bzr('commit --unchanged --message f')
        self.run_bzr('branch . ../branch2')
        self.run_bzr('branch . ../branch3')
        self.run_bzr('commit --unchanged --message peter')
        os.chdir('../branch2')
        self.run_bzr('merge ../branch1')
        self.run_bzr('commit --unchanged --message pumpkin')
        os.chdir('../branch3')
        self.run_bzr('merge ../branch2')
        message = self.run_bzr('status')[0]

    def test_conflicts(self):
        """Handling of merge conflicts"""
        self.create_conflicts()
        self.run_bzr('merge ../other --show-base', retcode=1)
        with open('hello') as f:
            conflict_text = f.read()
        self.assertTrue('<<<<<<<' in conflict_text)
        self.assertTrue('>>>>>>>' in conflict_text)
        self.assertTrue('=======' in conflict_text)
        self.assertTrue('|||||||' in conflict_text)
        self.assertTrue('hi world' in conflict_text)
        self.run_bzr('revert')
        self.run_bzr('resolve --all')
        self.run_bzr('merge ../other', retcode=1)
        with open('hello') as f:
            conflict_text = f.read()
        self.assertTrue('|||||||' not in conflict_text)
        self.assertTrue('hi world' not in conflict_text)
        result = self.run_bzr('conflicts')[0]
        self.assertEqual(result, "Text conflict in hello\nText conflict in"
                         " question\n")
        result = self.run_bzr('status')[0]
        self.assertTrue("conflicts:\n  Text conflict in hello\n"
                        "  Text conflict in question\n" in result, result)
        self.run_bzr('resolve hello')
        result = self.run_bzr('conflicts')[0]
        self.assertEqual(result, "Text conflict in question\n")
        self.run_bzr('commit -m conflicts', retcode=3)
        self.run_bzr('resolve --all')
        result = self.run_bzr('conflicts')[0]
        self.run_bzr('commit -m conflicts')
        self.assertEqual(result, "")

    def test_push(self):
        # create a source branch
        os.mkdir('my-branch')
        os.chdir('my-branch')
        self.example_branch()

        # with no push target, fail
        self.run_bzr('push', retcode=3)
        # with an explicit target work
        self.run_bzr('push ../output-branch')
        # with an implicit target work
        self.run_bzr('push')
        # nothing missing
        self.run_bzr('missing ../output-branch')
        # advance this branch
        self.run_bzr('commit --unchanged -m unchanged')

        os.chdir('../output-branch')
        # There is no longer a difference as long as we have
        # access to the working tree
        self.run_bzr('diff')

        # But we should be missing a revision
        self.run_bzr('missing ../my-branch', retcode=1)

        # diverge the branches
        self.run_bzr('commit --unchanged -m unchanged')
        os.chdir('../my-branch')
        # cannot push now
        self.run_bzr('push', retcode=3)
        # and there are difference
        self.run_bzr('missing ../output-branch', retcode=1)
        self.run_bzr('missing --verbose ../output-branch', retcode=1)
        # but we can force a push
        self.run_bzr('push --overwrite')
        # nothing missing
        self.run_bzr('missing ../output-branch')

        # pushing to a new dir with no parent should fail
        self.run_bzr('push ../missing/new-branch', retcode=3)
        # unless we provide --create-prefix
        self.run_bzr('push --create-prefix ../missing/new-branch')
        # nothing missing
        self.run_bzr('missing ../missing/new-branch')

    def test_external_command(self):
        """Test that external commands can be run by setting the path
        """
        # We don't at present run brz in a subprocess for blackbox tests, and so
        # don't really capture stdout, only the internal python stream.
        # Therefore we don't use a subcommand that produces any output or does
        # anything -- we just check that it can be run successfully.
        cmd_name = 'test-command'
        if sys.platform == 'win32':
            cmd_name += '.bat'
        self.overrideEnv('BZRPATH', None)

        f = open(cmd_name, 'wb')
        if sys.platform == 'win32':
            f.write(b'@echo off\n')
        else:
            f.write(b'#!/bin/sh\n')
        # f.write('echo Hello from test-command')
        f.close()
        os.chmod(cmd_name, 0o755)

        # It should not find the command in the local
        # directory by default, since it is not in my path
        self.run_bzr(cmd_name, retcode=3)

        # Now put it into my path
        self.overrideEnv('BZRPATH', '.')
        self.run_bzr(cmd_name)

        # Make sure empty path elements are ignored
        self.overrideEnv('BZRPATH', os.pathsep)
        self.run_bzr(cmd_name, retcode=3)


def listdir_sorted(dir):
    L = sorted(os.listdir(dir))
    return L


class OldTests(TestCaseWithTransport):
    """old tests moved from ./testbzr."""

    def test_bzr(self):
        from os import chdir, mkdir
        from os.path import exists

        progress = self.log

        progress("basic branch creation")
        mkdir('branch1')
        chdir('branch1')
        self.run_bzr('init')

        self.assertIsSameRealPath(self.run_bzr('root')[0].rstrip(),
                                  osutils.pathjoin(self.test_dir, 'branch1'))

        progress("status of new file")

        with open('test.txt', 'w') as f:
            f.write('hello world!\n')

        self.assertEqual(self.run_bzr('unknowns')[0], 'test.txt\n')

        out = self.run_bzr("status")[0]
        self.assertEqual(out, 'unknown:\n  test.txt\n')

        with open('test2.txt', 'w') as f:
            f.write('goodbye cruel world...\n')

        out = self.run_bzr("status test.txt")[0]
        self.assertEqual(out, "unknown:\n  test.txt\n")

        out = self.run_bzr("status")[0]
        self.assertEqual(out, ("unknown:\n" "  test.txt\n" "  test2.txt\n"))

        os.unlink('test2.txt')

        progress("command aliases")
        out = self.run_bzr("st")[0]
        self.assertEqual(out, ("unknown:\n" "  test.txt\n"))

        out = self.run_bzr("stat")[0]
        self.assertEqual(out, ("unknown:\n" "  test.txt\n"))

        progress("command help")
        self.run_bzr("help st")
        self.run_bzr("help")
        self.run_bzr("help commands")
        self.run_bzr("help slartibartfast", retcode=3)

        out = self.run_bzr("help ci")[0]
        out.index('Aliases:  ci, checkin\n')

        with open('hello.txt', 'w') as f:
            f.write('some nice new content\n')

        self.run_bzr("add hello.txt")

        with open('msg.tmp', 'w') as f:
            f.write('this is my new commit\nand it has multiple lines, for fun')

        self.run_bzr('commit -F msg.tmp')

        self.assertEqual(self.run_bzr('revno')[0], '1\n')
        self.run_bzr('export -r 1 export-1.tmp')
        self.run_bzr('export export.tmp')

        self.run_bzr('log')
        self.run_bzr('log -v')
        self.run_bzr('log -v --forward')
        self.run_bzr('log -m', retcode=3)
        log_out = self.run_bzr('log -m commit')[0]
        self.assertTrue("this is my new commit\n  and" in log_out)
        self.assertTrue("rename nested" not in log_out)
        self.assertTrue('revision-id' not in log_out)
        self.assertTrue(
            'revision-id' in self.run_bzr('log --show-ids -m commit')[0])

        log_out = self.run_bzr('log --line')[0]
        # determine the widest line we want
        max_width = osutils.terminal_width()
        if max_width is not None:
            for line in log_out.splitlines():
                self.assertTrue(len(line) <= max_width - 1, len(line))
        self.assertTrue("this is my new commit and" not in log_out)
        self.assertTrue("this is my new commit" in log_out)

        progress("file with spaces in name")
        mkdir('sub directory')
        with open('sub directory/file with spaces ', 'w') as f:
            f.write('see how this works\n')
        self.run_bzr('add .')
        self.run_bzr('diff', retcode=1)
        self.run_bzr('commit -m add-spaces')
        self.run_bzr('check')

        self.run_bzr('log')
        self.run_bzr('log --forward')

        self.run_bzr('info')

        if osutils.supports_symlinks(self.test_dir):
            progress("symlinks")
            mkdir('symlinks')
            chdir('symlinks')
            self.run_bzr('init')
            os.symlink("NOWHERE1", "link1")
            self.run_bzr('add link1')
            self.assertEqual(self.run_bzr('unknowns')[0], '')
            self.run_bzr(['commit', '-m', '1: added symlink link1'])

            mkdir('d1')
            self.run_bzr('add d1')
            self.assertEqual(self.run_bzr('unknowns')[0], '')
            os.symlink("NOWHERE2", "d1/link2")
            self.assertEqual(self.run_bzr('unknowns')[0], 'd1/link2\n')
            # is d1/link2 found when adding d1
            self.run_bzr('add d1')
            self.assertEqual(self.run_bzr('unknowns')[0], '')
            os.symlink("NOWHERE3", "d1/link3")
            self.assertEqual(self.run_bzr('unknowns')[0], 'd1/link3\n')
            self.run_bzr(['commit', '-m', '2: added dir, symlink'])

            self.run_bzr('rename d1 d2')
            self.run_bzr('move d2/link2 .')
            self.run_bzr('move link1 d2')
            self.assertEqual(os.readlink("./link2"), "NOWHERE2")
            self.assertEqual(os.readlink("d2/link1"), "NOWHERE1")
            self.run_bzr('add d2/link3')
            self.run_bzr('diff', retcode=1)
            self.run_bzr(['commit', '-m',
                          '3: rename of dir, move symlinks, add link3'])

            os.unlink("link2")
            os.symlink("TARGET 2", "link2")
            os.unlink("d2/link1")
            os.symlink("TARGET 1", "d2/link1")
            self.run_bzr('diff', retcode=1)
            self.assertEqual(self.run_bzr("relpath d2/link1")[0], "d2/link1\n")
            self.run_bzr(['commit', '-m', '4: retarget of two links'])

            self.run_bzr('remove --keep d2/link1')
            self.assertEqual(self.run_bzr('unknowns')[0], 'd2/link1\n')
            self.run_bzr(['commit', '-m', '5: remove d2/link1'])
            # try with the rm alias
            self.run_bzr('add d2/link1')
            self.run_bzr(['commit', '-m', '6: add d2/link1'])
            self.run_bzr('rm --keep d2/link1')
            self.assertEqual(self.run_bzr('unknowns')[0], 'd2/link1\n')
            self.run_bzr(['commit', '-m', '7: remove d2/link1'])

            os.mkdir("d1")
            self.run_bzr('add d1')
            self.run_bzr('rename d2/link3 d1/link3new')
            self.assertEqual(self.run_bzr('unknowns')[0], 'd2/link1\n')
            self.run_bzr(['commit', '-m',
                          '8: remove d2/link1, move/rename link3'])

            self.run_bzr('check')

            self.run_bzr('export -r 1 exp1.tmp')
            chdir("exp1.tmp")
            self.assertEqual(listdir_sorted("."), ["link1"])
            self.assertEqual(os.readlink("link1"), "NOWHERE1")
            chdir("..")

            self.run_bzr('export -r 2 exp2.tmp')
            chdir("exp2.tmp")
            self.assertEqual(listdir_sorted("."), ["d1", "link1"])
            chdir("..")

            self.run_bzr('export -r 3 exp3.tmp')
            chdir("exp3.tmp")
            self.assertEqual(listdir_sorted("."), ["d2", "link2"])
            self.assertEqual(listdir_sorted("d2"), ["link1", "link3"])
            self.assertEqual(os.readlink("d2/link1"), "NOWHERE1")
            self.assertEqual(os.readlink("link2"), "NOWHERE2")
            chdir("..")

            self.run_bzr('export -r 4 exp4.tmp')
            chdir("exp4.tmp")
            self.assertEqual(listdir_sorted("."), ["d2", "link2"])
            self.assertEqual(os.readlink("d2/link1"), "TARGET 1")
            self.assertEqual(os.readlink("link2"), "TARGET 2")
            self.assertEqual(listdir_sorted("d2"), ["link1", "link3"])
            chdir("..")

            self.run_bzr('export -r 5 exp5.tmp')
            chdir("exp5.tmp")
            self.assertEqual(listdir_sorted("."), ["d2", "link2"])
            self.assertTrue(os.path.islink("link2"))
            self.assertTrue(listdir_sorted("d2") == ["link3"])
            chdir("..")

            self.run_bzr('export -r 8 exp6.tmp')
            chdir("exp6.tmp")
            self.assertEqual(listdir_sorted("."), ["d1", "d2", "link2"])
            self.assertEqual(listdir_sorted("d1"), ["link3new"])
            self.assertEqual(listdir_sorted("d2"), [])
            self.assertEqual(os.readlink("d1/link3new"), "NOWHERE3")
            chdir("..")
        else:
            progress("skipping symlink tests")


class RemoteTests:
    """Test brz ui commands against remote branches."""

    def test_branch(self):
        os.mkdir('from')
        wt = self.make_branch_and_tree('from')
        branch = wt.branch
        wt.commit('empty commit for nonsense', allow_pointless=True)
        url = self.get_readonly_url('from')
        self.run_bzr(['branch', url, 'to'])
        branch = Branch.open('to')
        self.assertEqual(1, branch.last_revision_info()[0])
        # the branch should be set in to to from
        self.assertEqual(url + '/', branch.get_parent())

    def test_log(self):
        self.build_tree(['branch/', 'branch/file'])
        self.run_bzr('init branch')[0]
        self.run_bzr('add branch/file')[0]
        self.run_bzr('commit -m foo branch')[0]
        url = self.get_readonly_url('branch/file')
        output = self.run_bzr('log %s' % url)[0]
        self.assertEqual(8, len(output.split('\n')))

    def test_check(self):
        self.build_tree(['branch/', 'branch/file'])
        self.run_bzr('init branch')[0]
        self.run_bzr('add branch/file')[0]
        self.run_bzr('commit -m foo branch')[0]
        url = self.get_readonly_url('branch/')
        self.run_bzr(['check', url])

    def test_push(self):
        # create a source branch
        os.mkdir('my-branch')
        os.chdir('my-branch')
        self.run_bzr('init')
        with open('hello', 'w') as f:
            f.write('foo')
        self.run_bzr('add hello')
        self.run_bzr('commit -m setup')

        # with an explicit target work
        self.run_bzr(['push', self.get_url('output-branch')])


class HTTPTests(TestCaseWithWebserver, RemoteTests):
    """Test various commands against a HTTP server."""


class SFTPTestsAbsolute(TestCaseWithSFTPServer, RemoteTests):
    """Test various commands against a SFTP server using abs paths."""


class SFTPTestsAbsoluteSibling(TestCaseWithSFTPServer, RemoteTests):
    """Test various commands against a SFTP server using abs paths."""

    def setUp(self):
        super().setUp()
        self._override_home = '/dev/noone/runs/tests/here'


class SFTPTestsRelative(TestCaseWithSFTPServer, RemoteTests):
    """Test various commands against a SFTP server using homedir rel paths."""

    def setUp(self):
        super().setUp()
        self._get_remote_is_absolute = False



# Copyright (C) 2007 Canonical Ltd
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

"""Tests for interface conformance of 'WorkingTree.add'"""

from breezy import errors, tests
from breezy.bzr import inventory
from breezy.tests.matchers import HasLayout, HasPathRelations
from breezy.tests.per_workingtree import TestCaseWithWorkingTree


class TestAdd(TestCaseWithWorkingTree):

    def assertTreeLayout(self, expected, tree):
        """Check that the tree has the correct layout."""
        self.assertThat(tree, HasLayout(expected))

    def assertPathRelations(self, previous_tree, tree, relations):
        self.assertThat(tree, HasPathRelations(previous_tree, relations))

    def test_add_one(self):
        tree = self.make_branch_and_tree('.')
        self.build_tree(['one'])
        tree.add('one')

        self.assertTreeLayout(['', 'one'], tree)

    def test_add_existing_id(self):
        """Adding an entry with a pre-existing id raises DuplicateFileId"""
        tree = self.make_branch_and_tree('.')
        if not tree.supports_setting_file_ids():
            self.skipTest("tree does not support setting file ids")
        self.build_tree(['a', 'b'])
        tree.add(['a'])
        self.assertRaises(
            inventory.DuplicateFileId, tree.add, ['b'], ids=[tree.path2id('a')])
        # And the entry should not have been added.
        self.assertTreeLayout(['', 'a'], tree)

    def test_add_old_id(self):
        """We can add an old id, as long as it doesn't exist now."""
        tree = self.make_branch_and_tree('.')
        if not tree.supports_setting_file_ids():
            self.skipTest("tree does not support setting file ids")
        self.build_tree(['a', 'b'])
        tree.add(['a'])
        file_id = tree.path2id('a')
        tree.commit('first')
        # And the entry should not have been added.
        tree.unversion(['a'])
        tree.add(['b'], ids=[file_id])
        self.assertPathRelations(
            tree.basis_tree(), tree,
            [('', ''), ('b', 'a')])

    def test_add_one_list(self):
        tree = self.make_branch_and_tree('.')
        self.build_tree(['one'])
        tree.add(['one'])

        self.assertTreeLayout(['', 'one'], tree)

    def test_add_one_new_id(self):
        tree = self.make_branch_and_tree('.')
        self.build_tree(['one'])
        tree.add(['one'])

        self.assertTreeLayout(['', 'one'], tree)

    def test_add_unicode(self):
        tree = self.make_branch_and_tree('.')
        try:
            self.build_tree(['f\xf6'])
        except UnicodeError:
            raise tests.TestSkipped('Filesystem does not support filename.')
        tree.add(['f\xf6'])

        self.assertTreeLayout(['', 'f\xf6'], tree)

    def test_add_subdir_with_ids(self):
        tree = self.make_branch_and_tree('.')
        if not tree.supports_setting_file_ids():
            self.skipTest("tree does not support setting file ids")
        self.build_tree(['dir/', 'dir/subdir/', 'dir/subdir/foo'])
        tree.add(['dir'], ids=[b'dir-id'])
        tree.add(['dir/subdir'], ids=[b'subdir-id'])
        tree.add(['dir/subdir/foo'], ids=[b'foo-id'])
        root_id = tree.path2id('')

        self.assertTreeLayout([('', root_id), ('dir/', b'dir-id'),
                               ('dir/subdir/', b'subdir-id'),
                               ('dir/subdir/foo', b'foo-id')], tree)

    def test_add_subdir(self):
        tree = self.make_branch_and_tree('.')
        self.build_tree(['dir/', 'dir/subdir/', 'dir/subdir/foo'])
        tree.add(['dir'])
        tree.add(['dir/subdir'])
        tree.add(['dir/subdir/foo'])

        self.assertTreeLayout(
            ['', 'dir/', 'dir/subdir/', 'dir/subdir/foo'], tree)

    def test_add_multiple(self):
        tree = self.make_branch_and_tree('.')
        self.build_tree(['a', 'b', 'dir/', 'dir/subdir/', 'dir/subdir/foo'])
        tree.add(['a', 'b', 'dir', 'dir/subdir', 'dir/subdir/foo'])

        self.assertTreeLayout(
            ['', 'a', 'b', 'dir/', 'dir/subdir/', 'dir/subdir/foo'],
            tree)

    def test_add_multiple_with_file_ids(self):
        tree = self.make_branch_and_tree('.')
        if not tree.supports_setting_file_ids():
            self.skipTest("tree does not support setting file ids")
        self.build_tree(['a', 'b', 'dir/', 'dir/subdir/', 'dir/subdir/foo'])
        tree.add(['a', 'b', 'dir', 'dir/subdir', 'dir/subdir/foo'],
                 ids=[b'a-id', b'b-id', b'dir-id', b'subdir-id', b'foo-id'])

        self.assertTreeLayout([('', tree.path2id('')), ('a', b'a-id'), ('b', b'b-id'),
                               ('dir/', b'dir-id'), ('dir/subdir/', b'subdir-id'),
                               ('dir/subdir/foo', b'foo-id')], tree)

    def test_add_invalid(self):
        tree = self.make_branch_and_tree('.')
        if not tree._format.supports_versioned_directories:
            raise tests.TestNotApplicable(
                'format does not support versioned directories')
        self.build_tree(['dir/', 'dir/subdir/', 'dir/subdir/foo'])

        self.assertRaises(errors.NotVersionedError,
                          tree.add, ['dir/subdir'])
        self.assertTreeLayout([''], tree)

    def test_add_after_remove(self):
        tree = self.make_branch_and_tree('.')
        if not tree._format.supports_versioned_directories:
            raise tests.TestNotApplicable(
                'format does not support versioned directories')
        self.build_tree(['dir/', 'dir/subdir/', 'dir/subdir/foo'])
        tree.add(['dir'])
        tree.commit('dir')
        tree.unversion(['dir'])
        self.assertRaises(errors.NotVersionedError,
                          tree.add, ['dir/subdir/foo'])

    def test_add_root(self):
        # adding the root should be a no-op, or at least not
        # do anything whacky.
        tree = self.make_branch_and_tree('.')
        with tree.lock_write():
            tree.add('')
            self.assertEqual([''], list(tree.all_versioned_paths()))
            # the root should have been changed to be a new unique root.
            if tree._format.supports_setting_file_ids:
                self.assertNotEqual(inventory.ROOT_ID, tree.path2id(''))

    def test_add_previously_added(self):
        # adding a path that was previously added should work
        tree = self.make_branch_and_tree('.')
        self.build_tree(['foo'])
        tree.add(['foo'])
        tree.unversion(['foo'])
        tree.add(['foo'])
        self.assertTrue(tree.has_filename('foo'))

    def test_add_previously_added_with_file_id(self):
        # adding a path that was previously added should work
        tree = self.make_branch_and_tree('.')
        if not tree.supports_setting_file_ids():
            self.skipTest("tree does not support setting file ids")
        self.build_tree(['foo'])
        tree.add(['foo'], ids=[b'foo-id'])
        tree.unversion(['foo'])
        tree.add(['foo'], ids=[b'foo-id'])
        self.assertEqual(b'foo-id', tree.path2id('foo'))

    def test_add_present_in_basis(self):
        # adding a path that was present in the basis should work.
        tree = self.make_branch_and_tree('.')
        self.build_tree(['foo'])
        tree.add(['foo'])
        tree.commit('add foo')
        tree.unversion(['foo'])
        tree.add(['foo'])
        self.assertTrue(tree.has_filename('foo'))

    def test_add_present_in_basis_with_file_ids(self):
        # adding a path that was present in the basis should work.
        tree = self.make_branch_and_tree('.')
        if not tree.supports_setting_file_ids():
            self.skipTest("tree does not support setting file ids")
        self.build_tree(['foo'])
        tree.add(['foo'], ids=[b'foo-id'])
        tree.commit('add foo')
        tree.unversion(['foo'])
        tree.add(['foo'], ids=[b'foo-id'])
        self.assertEqual(b'foo-id', tree.path2id('foo'))



# Copyright (C) 2005-2012 Canonical Ltd
# Copyright (C) 2020 Breezy Developers
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

import fastbencode as bencode

from .. import errors, trace
from .. import transport as _mod_transport
from ..tag import Tags


class BasicTags(Tags):
    """Tag storage in an unversioned branch control file.
    """

    def set_tag(self, tag_name, tag_target):
        """Add a tag definition to the branch.

        Behaviour if the tag is already present is not defined (yet).
        """
        # all done with a write lock held, so this looks atomic
        with self.branch.lock_write():
            master = self.branch.get_master_branch()
            if master is not None:
                master.tags.set_tag(tag_name, tag_target)
            td = self.get_tag_dict()
            td[tag_name] = tag_target
            self._set_tag_dict(td)

    def lookup_tag(self, tag_name):
        """Return the referent string of a tag"""
        td = self.get_tag_dict()
        try:
            return td[tag_name]
        except KeyError:
            raise errors.NoSuchTag(tag_name)

    def get_tag_dict(self):
        with self.branch.lock_read():
            try:
                tag_content = self.branch._get_tags_bytes()
            except _mod_transport.NoSuchFile:
                # ugly, but only abentley should see this :)
                trace.warning('No branch/tags file in %s.  '
                              'This branch was probably created by bzr 0.15pre.  '
                              'Create an empty file to silence this message.'
                              % (self.branch, ))
                return {}
            return self._deserialize_tag_dict(tag_content)

    def delete_tag(self, tag_name):
        """Delete a tag definition.
        """
        with self.branch.lock_write():
            d = self.get_tag_dict()
            try:
                del d[tag_name]
            except KeyError:
                raise errors.NoSuchTag(tag_name)
            master = self.branch.get_master_branch()
            if master is not None:
                try:
                    master.tags.delete_tag(tag_name)
                except errors.NoSuchTag:
                    pass
            self._set_tag_dict(d)

    def _set_tag_dict(self, new_dict):
        """Replace all tag definitions

        WARNING: Calling this on an unlocked branch will lock it, and will
        replace the tags without warning on conflicts.

        :param new_dict: Dictionary from tag name to target.
        """
        return self.branch._set_tags_bytes(self._serialize_tag_dict(new_dict))

    def _serialize_tag_dict(self, tag_dict):
        td = {k.encode('utf-8'): v
                  for k, v in tag_dict.items()}
        return bencode.bencode(td)

    def _deserialize_tag_dict(self, tag_content):
        """Convert the tag file into a dictionary of tags"""
        # was a special case to make initialization easy, an empty definition
        # is an empty dictionary
        if tag_content == b'':
            return {}
        try:
            r = {}
            for k, v in bencode.bdecode(tag_content).items():
                r[k.decode('utf-8')] = v
            return r
        except ValueError as e:
            raise ValueError("failed to deserialize tag dictionary %r: %s"
                             % (tag_content, e))



# Copyright (C) 2009-2018 Jelmer Vernooij <jelmer@jelmer.uk>

# Based on the original from bzr-svn:
# Copyright (C) 2009 Lukas Lalinsky <lalinsky@gmail.com>
# Copyright (C) 2009 Jelmer Vernooij <jelmer@samba.org>

# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

"""Support in "brz send" for git-am style patches."""

import time
from io import BytesIO

from dulwich import __version__ as dulwich_version
from dulwich.objects import Blob

from .. import __version__ as brz_version
from .. import branch as _mod_branch
from .. import diff as _mod_diff
from .. import errors, osutils
from .. import revision as _mod_revision
from ..merge_directive import BaseMergeDirective
from .mapping import object_mode
from .object_store import get_object_store

version_tail = "Breezy %s, dulwich %d.%d.%d" % (
    (brz_version, ) + dulwich_version[:3])


class GitDiffTree(_mod_diff.DiffTree):
    """Provides a text representation between two trees, formatted for svn."""

    def _show_diff(self, specific_files, extra_trees):
        from dulwich.patch import write_blob_diff
        iterator = self.new_tree.iter_changes(
            self.old_tree, specific_files=specific_files,
            extra_trees=extra_trees, require_versioned=True)
        has_changes = 0

        def get_encoded_path(path):
            if path is not None:
                return path.encode(self.path_encoding, "replace")

        def get_file_mode(tree, path, kind, executable):
            if path is None:
                return 0
            return object_mode(kind, executable)

        def get_blob(present, tree, path):
            if present:
                with tree.get_file(path) as f:
                    return Blob.from_string(f.read())
            else:
                return None
        trees = (self.old_tree, self.new_tree)
        for change in iterator:
            # The root does not get diffed, and items with no known kind (that
            # is, missing) in both trees are skipped as well.
            if change.parent_id == (None, None) or change.kind == (None, None):
                continue
            path_encoded = (get_encoded_path(change.path[0]),
                            get_encoded_path(change.path[1]))
            present = ((change.kind[0] not in (None, 'directory')),
                       (change.kind[1] not in (None, 'directory')))
            if not present[0] and not present[1]:
                continue
            contents = (get_blob(present[0], trees[0], change.path[0]),
                        get_blob(present[1], trees[1], change.path[1]))
            renamed = (change.parent_id[0], change.name[0]) != (change.parent_id[1], change.name[1])
            mode = (get_file_mode(trees[0], path_encoded[0],
                                  change.kind[0], change.executable[0]),
                    get_file_mode(trees[1], path_encoded[1],
                                  change.kind[1], change.executable[1]))
            write_blob_diff(self.to_file,
                            (path_encoded[0], mode[0], contents[0]),
                            (path_encoded[1], mode[1], contents[1]))
            has_changes |= (change.changed_content or renamed)
        return has_changes


def generate_patch_filename(num, summary):
    return "%04d-%s.patch" % (num, summary.replace("/", "_").rstrip("."))


class GitMergeDirective(BaseMergeDirective):

    multiple_output_files = True

    def __init__(self, revision_id, testament_sha1, time, timezone,
                 target_branch, source_branch=None, message=None,
                 patches=None, local_target_branch=None):
        super().__init__(
            revision_id=revision_id, testament_sha1=testament_sha1, time=time,
            timezone=timezone, target_branch=target_branch, patch=None,
            source_branch=source_branch, message=message, bundle=None)
        self.patches = patches

    def to_lines(self):
        return self.patch.splitlines(True)

    def to_files(self):
        return ((summary, patch.splitlines(True)) for (summary, patch) in self.patches)

    @classmethod
    def _generate_commit(cls, repository, revision_id, num, total,
                         context=_mod_diff.DEFAULT_CONTEXT_AMOUNT):
        s = BytesIO()
        store = get_object_store(repository)
        with store.lock_read():
            commit = store[repository.lookup_bzr_revision_id(revision_id)[0]]
        from dulwich.patch import get_summary, write_commit_patch
        try:
            lhs_parent = repository.get_revision(revision_id).parent_ids[0]
        except IndexError:
            lhs_parent = _mod_revision.NULL_REVISION
        tree_1 = repository.revision_tree(lhs_parent)
        tree_2 = repository.revision_tree(revision_id)
        contents = BytesIO()
        differ = GitDiffTree.from_trees_options(
            tree_1, tree_2, contents, 'utf8', None, 'a/', 'b/', None,
            context_lines=context)
        differ.show_diff(None, None)
        write_commit_patch(s, commit, contents.getvalue(), (num, total),
                           version_tail)
        summary = generate_patch_filename(num, get_summary(commit))
        return summary, s.getvalue()

    @classmethod
    def from_objects(cls, repository, revision_id, time, timezone,
                     target_branch, local_target_branch=None,
                     public_branch=None, message=None):
        patches = []
        submit_branch = _mod_branch.Branch.open(target_branch)
        with submit_branch.lock_read():
            submit_revision_id = submit_branch.last_revision()
            repository.fetch(submit_branch.repository, submit_revision_id)
            graph = repository.get_graph()
            todo = graph.find_difference(submit_revision_id, revision_id)[1]
            total = len(todo)
            for i, revid in enumerate(graph.iter_topo_order(todo)):
                patches.append(cls._generate_commit(repository, revid, i + 1,
                                                    total))
        return cls(revision_id, None, time, timezone,
                   target_branch=target_branch, source_branch=public_branch,
                   message=message, patches=patches)


def send_git(branch, revision_id, submit_branch, public_branch, no_patch,
             no_bundle, message, base_revision_id, local_target_branch=None):
    if no_patch:
        raise errors.CommandError(
            "no patch not supported for git-am style patches")
    if no_bundle:
        raise errors.CommandError(
            "no bundle not supported for git-am style patches")
    return GitMergeDirective.from_objects(
        branch.repository, revision_id, time.time(),
        osutils.local_time_offset(), submit_branch,
        public_branch=public_branch, message=message,
        local_target_branch=local_target_branch)



# Copyright (C) 2011 Canonical Ltd
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA


from ... import errors, urlutils
from ...commands import Command
from ...controldir import ControlDir
from ...option import Option


class cmd_fetch_all_records(Command):
    __doc__ = """Fetch all records from another repository.

    This inserts every key from SOURCE_REPO into the target repository.  Unlike
    regular fetches this doesn't assume any relationship between keys (e.g.
    that text X may be assumed to be present if inventory Y is present), so it
    can be used to repair repositories where invariants about those
    relationships have somehow been violated.
    """

    hidden = True
    takes_args = ['source_repo']
    takes_options = [
        'directory',
        Option('dry-run',
               help="Show what would be done, but don't actually do anything."),
        ]

    def run(self, source_repo, directory='.', dry_run=False):
        try:
            source = ControlDir.open(source_repo).open_repository()
        except (errors.NotBranchError, urlutils.InvalidURL):
            print("Not a branch or invalid URL: %s" % source_repo,
                  file=self.outf)
            return

        try:
            target = ControlDir.open(directory).open_repository()
        except (errors.NotBranchError, urlutils.InvalidURL):
            print("Not a branch or invalid URL: %s" %
                  directory, file=self.outf)
            return

        self.add_cleanup(source.lock_read().unlock)
        self.add_cleanup(target.lock_write().unlock)

        # We need to find the keys to insert before we start the stream.
        # Otherwise we'll be querying the target repo while we're trying to
        # insert into it.
        needed = []
        for vf_name in ['signatures', 'texts', 'chk_bytes', 'inventories',
                        'revisions']:
            vf = getattr(source, vf_name)
            target_vf = getattr(target, vf_name)
            source_keys = vf.keys()
            target_keys = target_vf.keys()
            keys = source_keys.difference(target_keys)
            needed.append((vf_name, keys))

        def source_stream():
            for vf_name, keys in needed:
                vf = getattr(source, vf_name)
                yield (vf_name, vf.get_record_stream(keys, 'unordered', True))

        resume_tokens, missing_keys = target._get_sink().insert_stream(
            source_stream(), source._format, [])

        if not resume_tokens:
            print("Done.", file=self.outf)
        else:
            print("Missing keys!", missing_keys, file=self.outf)
